#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Base classes for PDDL parsers."""

import sys
from abc import ABC
from typing import Any, Type, TypeVar, cast

from lark import Lark, Transformer

from pddl.parser import GRAMMAR_FILE, PARSERS_DIRECTORY

T = TypeVar("T")


class BaseParser(Transformer[Any, T], ABC):
    """Base class for PDDL parsers."""

    start_symbol: str
    transformer_cls: Type[Transformer[Any, T]]

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the parser with the given grammar file."""
        super().__init__(*args, **kwargs)
        self._transformer = self.transformer_cls()
        self._parser = Lark(
            GRAMMAR_FILE.read_text(),
            parser="lalr",
            import_paths=[PARSERS_DIRECTORY],
            start=self.start_symbol,
            transformer=self._transformer,
        )

    def __call__(self, text: str) -> T:
        """Call."""
        return self._call_parser(text, self._parser)

    @classmethod
    def _call_parser(cls, text: str, parser: Lark) -> T:
        """
        Parse a text with a Lark parser and transformer.

        To produce a better traceback in case of an error, the function will temporarily overwrite the
        sys.tracebacklimit value of the current interpreter.

        :param text: the text to parse
        :param parser: the Lark parser object
        :param transformer: the Lark transformer object
        :return: the object returned by the parser
        """
        old_tracebacklimit = getattr(sys, "tracebacklimit", None)
        try:
            sys.tracebacklimit = 0  # noqa
            result = cast(T, parser.parse(text))
            sys.tracebacklimit = None  # type: ignore
        finally:
            if old_tracebacklimit is not None:
                sys.tracebacklimit = old_tracebacklimit
        return result



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Implementation of the PDDL domain parser."""

from typing import (
    AbstractSet,
    Any,
    Dict,
    FrozenSet,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
)

from lark import ParseError, Transformer

from pddl.action import Action
from pddl.core import Domain
from pddl.custom_types import name
from pddl.exceptions import PDDLMissingRequirementError, PDDLParsingError
from pddl.helpers.base import assert_
from pddl.logic.base import And, ExistsCondition, ForallCondition, Imply, Not, OneOf, Or
from pddl.logic.effects import Forall, When
from pddl.logic.functions import Assign, Decrease, Divide
from pddl.logic.functions import EqualTo as FunctionEqualTo
from pddl.logic.functions import (
    FunctionExpression,
    GreaterEqualThan,
    GreaterThan,
    Increase,
    LesserEqualThan,
    LesserThan,
    Minus,
    NumericFunction,
    NumericValue,
    Plus,
    ScaleDown,
    ScaleUp,
    Times,
)
from pddl.logic.predicates import DerivedPredicate, EqualTo, Predicate
from pddl.logic.terms import Constant, Variable
from pddl.parser._update_type_tags import update_type_tags
from pddl.parser.base import BaseParser
from pddl.parser.symbols import BINARY_COMP_SYMBOLS, Symbols
from pddl.parser.typed_list_parser import TypedListParser
from pddl.requirements import Requirements, _extend_domain_requirements


class DomainTransformer(Transformer[Any, Domain]):
    """Domain Transformer."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the domain transformer."""
        super().__init__(*args, **kwargs)

        self._constants_by_name: Dict[str, Constant] = {}
        self._predicates_by_name: Dict[str, Predicate] = {}
        self._functions_by_name: Dict[str, FunctionExpression] = {}
        self._current_parameters_by_name: Dict[str, Variable] = {}
        self._requirements: Set[Requirements] = set()
        self._extended_requirements: Set[Requirements] = set()
        self._types: Optional[Mapping[name, Optional[name]]] = None

    @property
    def types_hierarchy(self) -> Mapping[name, Optional[name]]:
        """Return the read-only types hierarchy."""
        if self._types is None:
            raise RuntimeError("Types hierarchy is not available yet.")
        return self._types

    def start(self, args):
        """Entry point."""
        return args[0]

    def domain(self, args):
        """Process the 'domain' rule."""
        args = [arg for arg in args if arg is not None]
        kwargs = {}
        actions = []
        derived_predicates = []
        for arg in args[2:-1]:
            if isinstance(arg, Action):
                actions.append(arg)
            elif isinstance(arg, DerivedPredicate):
                derived_predicates.append(arg)
            else:
                assert_(isinstance(arg, dict))
                kwargs.update(arg)
        kwargs.update(actions=actions, derived_predicates=derived_predicates)
        self._types = None
        return Domain(**kwargs)

    def domain_def(self, args):
        """Process the 'domain_def' rule."""
        return dict(name=args[2])

    def requirements(self, args):
        """Process the 'requirements' rule."""
        self._requirements = {Requirements(r[1:]) for r in args[2:-1]}
        self._extended_requirements = _extend_domain_requirements(self._requirements)

        return dict(requirements=self._requirements)

    def types(self, args):
        """Parse the 'types' rule."""
        has_typing_requirement = self._has_requirement(Requirements.TYPING)
        types_definition = args[2]
        have_type_hierarchy = any(types_definition.values())
        if have_type_hierarchy and not has_typing_requirement:
            raise PDDLMissingRequirementError(Requirements.TYPING)
        for k in types_definition:
            if types_definition[k] == Symbols.OBJECT.value:
                types_definition[k] = None

        # record types hierarchy for later use during the parsing
        # safety check: make sure self._types is not set
        assert_(self._types is None, "parser is in an unexpected state")
        self._types = types_definition

        return dict(types=types_definition)

    def constants(self, args):
        """Process the 'constant_def' rule."""
        self._constants_by_name = {
            name: Constant(name, type_tags) for name, type_tags in args[2].items()
        }
        return dict(constants=list(self._constants_by_name.values()))

    def predicates(self, args):
        """Process the 'predicates' rule."""
        predicates = args[2:-1]
        self._predicates_by_name = {p.name: p for p in predicates}
        return dict(predicates=predicates)

    def functions(self, args):
        """Process the 'functions' rule."""
        function_definition = args[2]
        # arg[2] is a dict with NumericFunction as keys and types as values, e.g., {(fuel ?l): number, (cost): number}
        return dict(functions=function_definition)

    def action_def(self, args):
        """Process the 'action_def' rule."""
        action_name = args[2]
        variables = args[4]

        # process action body
        _children = args[5].children
        action_body = {
            _children[i][1:]: _children[i + 1] for i in range(0, len(_children), 2)
        }
        return Action(action_name, variables, **action_body)

    def derived_predicates(self, args):
        """Process the 'derived_predicates' rule."""
        dp_predicate = args[2]
        dp_condition = args[3]

        # we need to make sure the variables in the condition of the derived predicate
        # have type tags compatible with the variables in the corresponding predicate definition
        # to do so, we read the type tags of the predicate arguments from the predicate definition in (:predicates ...)
        if dp_predicate.name not in self._predicates_by_name:
            raise ParseError(f"Predicate '{dp_predicate.name}' not defined.")
        predicate_def = self._predicates_by_name[dp_predicate.name]
        if len(predicate_def.terms) != len(dp_predicate.terms):
            raise ParseError(
                f"Derived predicate '{dp_predicate.name}' has {len(dp_predicate.terms)} arguments, "
                f"but predicate definition has {len(predicate_def.terms)} arguments."
            )

        # this is the mapping we will use to update the type tags of the variables occurrences
        # in the condition of the derived predicate
        current_var_to_types: Dict[str, FrozenSet[str]] = {}
        for idx, term in enumerate(dp_predicate.terms):
            assert_(
                isinstance(term, Variable),
                f"Expected Variable, got {type(term).__name__}",
            )

            # if the term has no type tags, we try to use the type tags from the predicate definition
            if len(term.type_tags) == 0:
                # if the predicate definition specifies type tags for this argument, we use it.
                # this has the effect of updating the type tags of the variable occurrence in both the dp_predicate
                # and the dp_condition.
                if len(predicate_def.terms[idx].type_tags) != 0:
                    current_var_to_types[term.name] = frozenset(
                        predicate_def.terms[idx].type_tags
                    )
            else:
                # otherwise, we use the type tags from the definition of the derived predicate
                current_var_to_types[term.name] = frozenset(term.type_tags)

                # however, in this case, we also need to make sure that the type tags in the condition of the
                # derived predicate are consistent with the type tags in the predicate definition
                try:
                    self._check_subtypes(
                        term.type_tags, predicate_def.terms[idx].type_tags
                    )
                except PDDLParsingError as e:
                    raise PDDLParsingError(
                        f"Type tags in condition and predicate definition do not match "
                        f"for variable '{term.name}' (position {idx}) in predicate '{dp_predicate.name}': {e}"
                    )

        new_predicate = update_type_tags(dp_predicate, current_var_to_types)
        new_condition = update_type_tags(dp_condition, current_var_to_types)
        return DerivedPredicate(new_predicate, new_condition)

    def action_parameters(self, args):
        """Process the 'action_parameters' rule."""
        self._current_parameters_by_name = {
            var_name: Variable(var_name, tags) for var_name, tags in args[1]
        }
        return list(self._current_parameters_by_name.values())

    def emptyor_pregd(self, args):
        """Process the 'emptyor_pregd' rule."""
        if len(args) == 2:
            return Or()
        else:
            assert_(len(args) == 1)
            return args[0]

    def gd_not(self, args):
        """Process the 'gd' not rule."""
        if not bool(
            {Requirements.NEG_PRECONDITION, Requirements.ADL}
            & self._extended_requirements
        ):
            # raise PDDLMissingRequirementError(Requirements.NEG_PRECONDITION)
            # TODO temporary change; remove
            pass
        return Not(args[2])

    def gd_and(self, args):
        """Process the 'gd_and' rule."""
        operands = args[2:-1]
        return And(*operands)

    def gd_or(self, args):
        """Process the 'gd' or rule."""
        if not bool(
            {Requirements.DIS_PRECONDITION, Requirements.ADL}
            & self._extended_requirements
        ):
            raise PDDLMissingRequirementError(Requirements.DIS_PRECONDITION)
        operands = args[2:-1]
        return Or(*operands)

    def gd_imply(self, args):
        """Process the 'gd' imply rule."""
        if not bool(
            {Requirements.DIS_PRECONDITION, Requirements.ADL}
            & self._extended_requirements
        ):
            raise PDDLMissingRequirementError(Requirements.DIS_PRECONDITION)
        return Imply(args[2], args[3])

    def gd_quantifiers(self, args):
        """Process the 'gd' quantifiers rule."""
        req, cond_class = {
            Symbols.FORALL.value: (
                Requirements.UNIVERSAL_PRECONDITION,
                ForallCondition,
            ),
            Symbols.EXISTS.value: (
                Requirements.EXISTENTIAL_PRECONDITION,
                ExistsCondition,
            ),
        }[args[1]]
        if not bool(
            {req, Requirements.QUANTIFIED_PRECONDITION, Requirements.ADL}
            & self._extended_requirements
        ):
            raise PDDLMissingRequirementError(req)
        variables = [Variable(var_name, tags) for var_name, tags in args[3]]
        condition = args[5]
        return cond_class(cond=condition, variables=variables)

    def gd_comparison(self, args):
        """Process the 'gd' comparison rule."""
        if not bool({Requirements.NUMERIC_FLUENTS, Requirements.FLUENTS}):
            raise PDDLMissingRequirementError(Requirements.NUMERIC_FLUENTS)
        left = args[2]
        right = args[3]
        if args[1] == Symbols.GREATER_EQUAL.value:
            return GreaterEqualThan(left, right)
        elif args[1] == Symbols.GREATER.value:
            return GreaterThan(left, right)
        elif args[1] == Symbols.LESSER_EQUAL.value:
            return LesserEqualThan(left, right)
        elif args[1] == Symbols.LESSER.value:
            return LesserThan(left, right)
        elif args[1] == Symbols.EQUAL.value:
            return FunctionEqualTo(left, right)
        else:
            raise PDDLParsingError(f"Unknown comparison operator: {args[1]}")

    def gd(self, args):
        """Process the 'gd' rule."""
        if len(args) == 1:
            return args[0]
        elif args[1] == Symbols.NOT.value:
            return self.gd_not(args)
        elif args[1] == Symbols.AND.value:
            return self.gd_and(args)
        elif args[1] == Symbols.OR.value:
            return self.gd_or(args)
        elif args[1] == Symbols.IMPLY.value:
            return self.gd_imply(args)
        elif args[1] in [Symbols.FORALL.value, Symbols.EXISTS.value]:
            return self.gd_quantifiers(args)
        elif args[1] in BINARY_COMP_SYMBOLS:
            return self.gd_comparison(args)

    def num_literal(self, args):
        """Process the 'num_literal' rule."""
        n = args[0]
        return float(n) if "." in n else int(n)

    def emptyor_effect(self, args):
        """Process the 'emptyor_effect' rule."""
        if len(args) == 2:
            return Or()
        else:
            return args[0]

    def effect(self, args):
        """Process the 'effect' rule."""
        if len(args) == 1:
            return args[0]
        if args[1] == Symbols.AND.value:
            return And(*args[2:-1])
        raise ValueError("case not recognized")

    def c_effect(self, args):
        """Process the 'c_effect' rule."""
        if len(args) == 1:
            return args[0]
        if args[1] == Symbols.FORALL.value:
            variables = [Variable(var_name, tags) for var_name, tags in args[3]]
            return Forall(effect=args[-2], variables=variables)
        if args[1] == Symbols.WHEN.value:
            return When(args[2], args[3])
        if args[1] == Symbols.ONEOF.value:
            if not bool({Requirements.NON_DETERMINISTIC} & self._extended_requirements):
                raise PDDLMissingRequirementError(Requirements.NON_DETERMINISTIC)
            return OneOf(*args[2:-1])
        raise ValueError()

    def p_effect(self, args):
        """Process the 'p_effect' rule."""
        if len(args) == 1:
            return args[0]
        else:
            return Not(args[2])

    def cond_effect(self, args):
        """Process the 'cond_effect' rule."""
        if len(args) >= 3 and args[1] == Symbols.AND.value:
            p_effects = args[2:-1]
            return And(*p_effects)
        assert_(len(args) == 1)
        return args[0]

    def num_effect(self, args):
        """Process the 'num_effect' rule."""
        if args[1] == Symbols.ASSIGN.value:
            return Assign(args[2], args[3])
        elif args[1] == Symbols.SCALE_UP.value:
            return ScaleUp(args[2], args[3])
        elif args[1] == Symbols.SCALE_DOWN.value:
            return ScaleDown(args[2], args[3])
        elif args[1] == Symbols.INCREASE.value:
            return Increase(args[2], args[3])
        elif args[1] == Symbols.DECREASE.value:
            return Decrease(args[2], args[3])
        else:
            raise PDDLParsingError(f"Unrecognized assign operator: {args[1]}")

    def _constant_or_variable(self, t):
        """Get the constant or variable with the given name."""
        # Case where the term is a free variable (bug) or comes from a parent quantifier
        if not isinstance(t, Constant) and t not in self._current_parameters_by_name:
            return Variable(str(t), {})
        return t if isinstance(t, Constant) else self._current_parameters_by_name[t]

    def atomic_formula_term(self, args):
        """Process the 'atomic_formula_term' rule."""
        if args[1] == Symbols.EQUAL.value:
            if not bool({Requirements.EQUALITY} & self._extended_requirements):
                raise PDDLMissingRequirementError(Requirements.EQUALITY)
            left = self._constant_or_variable(args[2])
            right = self._constant_or_variable(args[3])
            return EqualTo(left, right)
        else:
            predicate_name = args[1]
            terms = list(map(self._constant_or_variable, args[2:-1]))
            return Predicate(predicate_name, *terms)

    def constant(self, args):
        """Process the 'constant' rule."""
        assert_(len(args) == 1, "Unexpected parsing error.")
        constant = self._constants_by_name.get(args[0], None)
        if constant is None:
            raise ParseError(f"Constant '{args[0]}' not defined.")
        return constant

    def _formula_skeleton(self, args) -> Sequence[Variable]:
        """Process the '_formula_skeleton' rule."""
        variable_data: Tuple[Tuple[str, Set[str]], ...] = args[2]
        variables: List[Variable] = [
            Variable(var_name, tags) for var_name, tags in variable_data
        ]
        return variables

    def atomic_formula_skeleton(self, args):
        """Process the 'atomic_formula_skeleton' rule."""
        predicate_name = args[1]
        variables = self._formula_skeleton(args)
        return Predicate(predicate_name, *variables)

    def atomic_function_skeleton(self, args):
        """Process the 'atomic_function_skeleton' rule."""
        if args[1] == Symbols.TOTAL_COST.value:
            if not bool({Requirements.ACTION_COSTS}):
                raise PDDLMissingRequirementError(Requirements.ACTION_COSTS)
            return NumericFunction("total-cost")
        function_name = args[1]
        variables = self._formula_skeleton(args)
        return NumericFunction(function_name, *variables)

    def f_exp(self, args):
        """Process the 'f_exp' rule."""
        if len(args) == 1:
            if not isinstance(args[0], NumericFunction):
                return NumericValue(args[0])
            return args[0]
        op = None
        if args[1] == Symbols.MINUS.value:
            op = Minus
        if args[1] == Symbols.PLUS.value:
            op = Plus
        if args[1] == Symbols.TIMES.value:
            op = Times
        if args[1] == Symbols.DIVIDE.value:
            op = Divide
        return (
            op(*args[2:-1])
            if op is not None
            else PDDLParsingError("Operator not recognized")
        )

    def f_head(self, args):
        """Process the 'f_head' rule."""
        if len(args) == 1:
            return NumericFunction(args[0])
        function_name = args[1]
        variables = list(map(self._constant_or_variable, args[2:-1]))
        return NumericFunction(function_name, *variables)

    def typed_list_name(self, args) -> Dict[name, Optional[name]]:
        """Process the 'typed_list_name' rule."""
        try:
            types_index = TypedListParser.parse_typed_list(args)
            return types_index.get_typed_list_of_names()
        except ValueError as e:
            raise self._raise_typed_list_parsing_error(args, e) from e

    def typed_list_variable(self, args) -> Tuple[Tuple[name, Set[name]], ...]:
        """
        Process the 'typed_list_variable' rule.

        Return a dictionary with as keys the terms and as value a set of types for each name.

        :param args: the argument of this grammar rule
        :return: a typed list (variable), i.e. a mapping from variables to the supported types
        """
        try:
            types_index = TypedListParser.parse_typed_list(args, allow_duplicates=True)
            return types_index.get_typed_list_of_variables()
        except ValueError as e:
            raise self._raise_typed_list_parsing_error(args, e) from e

    def f_typed_list_atomic_function_skeleton(self, args):
        """Process the 'f_typed_list_atomic_function_skeleton' rule."""
        try:
            types_index = TypedListParser.parse_typed_list(args)
            return types_index.get_typed_list_of_names()
        except ValueError as e:
            raise self._raise_typed_list_parsing_error(args, e) from e

    def _raise_typed_list_parsing_error(self, args, exception) -> PDDLParsingError:
        string_list = [
            str(arg) if isinstance(arg, str) else list(map(str, arg)) for arg in args
        ]
        return PDDLParsingError(
            f"error while parsing tokens {string_list}: {str(exception)}"
        )

    def type_def(self, args):
        """Parse the 'type_def' rule."""
        assert_(len(args) != 0, "unexpected parser state: empty type_def")

        if len(args) == 1:
            # single-typed type-def, return
            return args

        # if we are here, type_def is of the form (either t1 ... tn)
        # ignore first and last tokens since they are brackets.
        either_keyword, types = args[1], args[2:-1]
        assert_(str(either_keyword) == Symbols.EITHER.value)
        return types

    def _has_requirement(self, requirement: Requirements) -> bool:
        """Check whether a requirement is satisfied by the current state of the domain parsing."""
        return requirement in self._extended_requirements

    def _check_subtypes(
        self, type_tags_left: AbstractSet[name], type_tags_right: AbstractSet[name]
    ):
        """Check that each type tag in the left set is a subtype of (one of) the type tags of the right."""
        if len(type_tags_right) == 0:
            # 'object' type includes everything
            return

        # for each left type, navigate the type hierarchy until we find a type in the right set, or fail
        for left_type in type_tags_left:
            current_type = left_type
            parent_type: Optional[name] = left_type
            while parent_type is not None and parent_type not in type_tags_right:
                parent_type = self.types_hierarchy.get(current_type, None)

            if parent_type is None:
                raise PDDLParsingError(
                    f"Type '{left_type}' not found in types: '{type_tags_right}'."
                )


class DomainParser(BaseParser[Domain]):
    """PDDL domain parser class."""

    transformer_cls = DomainTransformer
    start_symbol = "domain"



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains the definition of the PDDL action class."""

from typing import Optional, Sequence

from pddl.custom_types import namelike, parse_name
from pddl.helpers.base import _typed_parameters, ensure_sequence
from pddl.logic import Variable
from pddl.logic.base import Formula
from pddl.logic.terms import Term


class Action:
    """A class for the PDDL Action."""

    def __init__(
        self,
        name: namelike,
        parameters: Sequence[Term],
        precondition: Optional[Formula] = None,
        effect: Optional[Formula] = None,
    ):
        """
        Initialize the action.

        :param name: the action name.
        :param parameters: the action parameters.
        :param precondition: the action precondition.
        :param effect: the action effect.
        """
        self._name: str = parse_name(name)
        self._parameters: Sequence[Variable] = ensure_sequence(parameters)
        self._precondition = precondition
        self._effect = effect

    @property
    def name(self) -> str:
        """Get the name."""
        return self._name

    @property
    def parameters(self) -> Sequence[Variable]:
        """Get the parameters."""
        return self._parameters

    @property
    def terms(self) -> Sequence[Term]:
        """Get the terms."""
        return self.parameters

    @property
    def precondition(self) -> Optional[Formula]:
        """Get the precondition."""
        return self._precondition

    @property
    def effect(self) -> Optional[Formula]:
        """Get the effect."""
        return self._effect

    def instantiate(self, mapping: Sequence[Term]) -> "Action":
        """
        Instantiate the action by replacing its parameters according to the given mapping.

        :param mapping: a list of terms to replace the parameters.
        :return: the instantiated action.
        """
        if len(mapping) != len(self.parameters):
            raise ValueError(
                f"Mapping length {len(mapping)} does not match number of parameters {len(self.parameters)}."
            )

        instantiation = {param: const for param, const in zip(self.parameters, mapping)}

        return Action(
            name=self.name,
            parameters=mapping,
            precondition=(
                self.precondition.instantiate(instantiation)
                if self.precondition
                else None
            ),
            effect=self.effect.instantiate(instantiation) if self.effect else None,
        )

    def __str__(self):
        """Get the string."""
        operator_str = "(:action {0}\n".format(self.name)
        operator_str += f"    :parameters ({_typed_parameters(self.parameters)})\n"
        if self.precondition is not None:
            operator_str += f"    :precondition {str(self.precondition)}\n"
        if self.effect is not None:
            operator_str += f"    :effect {str(self.effect)}\n"
        operator_str += ")"
        return operator_str

    def __eq__(self, other):
        """Check equality between two Actions."""
        return (
            isinstance(other, Action)
            and self.name == other.name
            and self.parameters == other.parameters
            and self.precondition == other.precondition
            and self.effect == other.effect
        )

    def __hash__(self):
        """Get the hash."""
        return hash((self.name, self.parameters, self.precondition, self.effect))

    def __repr__(self) -> str:
        """Get an unambiguous string representation."""
        return (
            f"{type(self).__name__}({self.name}, parameters={', '.join(map(str, self.parameters))}, "
            f"precondition={self.precondition}, effect={self.effect})"
        )



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Utilities for docs tests."""

import traceback
from functools import partial
from pathlib import Path
from typing import Dict, List, Optional

import mistune
import mistune.renderers
import pytest

MISTUNE_BLOCK_CODE_ID = "block_code"


def compile_and_exec(code: str, locals_dict: Optional[Dict] = None) -> Dict:
    """
    Compile and exec the code.

    :param code: the code to execute.
    :param locals_dict: the dictionary of local variables.
    :return: the dictionary of locals.
    """
    locals_dict = {} if locals_dict is None else locals_dict
    try:
        code_obj = compile(code, "fakemodule", "exec")
        exec(code_obj, locals_dict)  # nosec
    except Exception:  # type: ignore
        pytest.fail(
            "The execution of the following code:\n{}\nfailed with error:\n{}".format(
                code, traceback.format_exc()
            )
        )
    return locals_dict


class BaseTestMarkdownDocs:
    """Base test class for blocksworld_fond Markdown documents."""

    MD_FILE: Optional[Path] = None
    code_blocks: List[Dict] = []

    @classmethod
    def setup_class(cls):
        """Set up class."""
        if cls.MD_FILE is None:
            raise ValueError("cannot set up method as MD_FILE is None")
        content = cls.MD_FILE.read_text()
        markdown_parser = mistune.create_markdown(renderer="ast")
        cls.blocks = markdown_parser(content)
        cls.code_blocks = list(filter(cls.block_code_filter, cls.blocks))

    @staticmethod
    def block_code_filter(block: Dict) -> bool:
        """Check Mistune block is a code block."""
        return block["type"] == MISTUNE_BLOCK_CODE_ID

    @staticmethod
    def type_filter(type_: Optional[str], b: Dict) -> bool:
        """
        Check Mistune code block is of a certain type.

        If the field "info" is None, return False.
        If type_ is None, this function always return true.

        :param type_: the expected type of block (optional)
        :param b: the block dicionary.
        :return: True if the block should be accepted, false otherwise.
        """
        if type_ is None:
            return True
        return b["type"].strip() == type_ if b["type"] is not None else False

    @classmethod
    def extract_code_blocks(cls, filter_: Optional[str] = None):
        """Extract code blocks from .md files."""
        actual_type_filter = partial(cls.type_filter, filter_)
        bash_code_blocks = filter(actual_type_filter, cls.code_blocks)
        return list(b["text"] for b in bash_code_blocks)


class BasePythonMarkdownDocs(BaseTestMarkdownDocs):
    """Test Markdown documentation by running Python snippets in sequence."""

    @classmethod
    def setup_class(cls):
        """
        Set up class.

        It sets the initial value of locals and globals.
        """
        super().setup_class()
        cls.locals = {}
        cls.globals = {}

    @classmethod
    def _python_selector(cls, block: Dict) -> bool:
        return block["type"] == MISTUNE_BLOCK_CODE_ID and (
            block["type"].strip() == "python" if block["type"] else False
        )

    def _assert(self, locals_, *mocks):
        """Do assertions after Python code execution."""

    def test_python_blocks(self, *mocks):
        """Run Python code block in sequence."""
        python_blocks = list(filter(self._python_selector, self.blocks))

        globals_, locals_ = self.globals, self.locals
        for python_block in python_blocks:
            python_code = python_block["text"]
            exec(python_code, globals_, locals_)  # nosec
        self._assert(locals_, *mocks)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This modules implements PDDL effects."""

import functools
from typing import AbstractSet, Collection, Optional, Union

from pddl.helpers.base import _typed_parameters, ensure_set
from pddl.helpers.cache_hash import cache_hash
from pddl.logic import Variable
from pddl.logic.base import And, Atomic, Formula, Not, OneOf
from pddl.parser.symbols import Symbols


@cache_hash
@functools.total_ordering
class When:
    """Represent the 'When' effect."""

    def __init__(self, condition: Formula, effect: "CondEffect") -> None:
        """
        Initialize the effect.

        :param condition: the condition
        :param effect: the effect
        """
        self._condition = condition
        self._effect = effect

    @property
    def condition(self) -> Formula:
        """Get the condition."""
        return self._condition

    @property
    def effect(self) -> "CondEffect":
        """Get the effect."""
        return self._effect

    def __str__(self) -> str:
        """Get the string representation."""
        return f"({Symbols.WHEN.value} {self._condition} {self.effect})"

    def __repr__(self) -> str:
        """Get an unambiguous string representation."""
        return f"{type(self).__name__}({self.condition}, {self.effect})"

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return (
            isinstance(other, type(self))
            and self.condition == other.condition
            and self.effect == other.effect
        )

    def __hash__(self) -> int:
        """Compute the hash of the object."""
        return hash((type(self), self.condition, self.effect))

    def __lt__(self, other):
        """Compare with another object."""
        if isinstance(other, When):
            return (self.condition, self.effect) < (other.condition, other.effect)
        return super().__lt__(other)


@cache_hash
@functools.total_ordering
class Forall:
    """Represent the 'Forall' effect."""

    def __init__(
        self, effect: "Effect", variables: Optional[Collection[Variable]] = None
    ) -> None:
        """Initialize the 'forall' effect."""
        self._effect = effect
        self._variables = ensure_set(variables)

    @property
    def effect(self) -> "Effect":
        """Get the effect."""
        return self._effect

    @property
    def variables(self) -> AbstractSet[Variable]:
        """Get the variables."""
        return self._variables

    def __str__(self) -> str:
        """Get the string representation."""
        return f"({Symbols.FORALL.value} ({_typed_parameters(self.variables)}) {self.effect})"

    def __repr__(self) -> str:
        """Get an unambiguous string representation."""
        return f"{type(self).__name__}({self.variables}, {self.effect})"

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return (
            isinstance(other, type(self))
            and self.variables == other.variables
            and self.effect == other.effect
        )

    def __hash__(self) -> int:
        """Compute the hash of the object."""
        return hash((type(self), self.variables, self.effect))

    def __lt__(self, other):
        """Compare with another object."""
        if isinstance(other, Forall):
            return (self.variables, self.effect) < (other.variables, other.effect)
        return super().__lt__(other)


PEffect = Union[Atomic, Not]
CEffect = Union[Forall, When, OneOf, "PEffect"]
Effect = Union[And, CEffect]
CondEffect = Union[And, "PEffect"]



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains the configurations for the tests."""

import inspect
import itertools
from pathlib import Path

import mistune
import pytest
from pytest_lazy_fixtures import lf

import pddl
from pddl.parser.domain import DomainParser
from pddl.parser.problem import ProblemParser
from pddl.parser.symbols import Symbols

_current_filepath = inspect.getframeinfo(inspect.currentframe()).filename  # type: ignore
TEST_DIRECTORY = Path(_current_filepath).absolute().parent
ROOT_DIRECTORY = TEST_DIRECTORY.parent
LIBRARY_DIRECTORY = ROOT_DIRECTORY / pddl.__name__
DOCS_DIRECTORY = ROOT_DIRECTORY / "docs"


FIXTURES_DIR = TEST_DIRECTORY / "fixtures"
FIXTURES_PDDL_FILES = FIXTURES_DIR / "pddl_files"
BLOCKSWORLD_FILES = FIXTURES_PDDL_FILES / "blocksworld-ipc08"
TRIANGLE_FILES = FIXTURES_PDDL_FILES / "triangle-tireworld"
BLOCKSWORLD_FOND_FILES = FIXTURES_PDDL_FILES / "blocksworld_fond"

DOMAIN_FILES = [*FIXTURES_PDDL_FILES.glob("./**/domain.pddl")]

DOMAIN_NAMES = [
    "acrobatics",
    "barman-sequential-optimal",
    "beam-walk",
    "blocksworld-ipc08",
    "blocksworld_fond",
    "cave-diving-sequential-optimal",
    # "depots-numeric-automatic",
    "doors",
    "drone",
    "earth_observation",
    "elevators",
    # "faults-ipc08",
    "first-responders-ipc08",
    "islands",
    "maintenance-sequential-satisficing-ipc2014",
    "miner",
    "rovers-numeric-automatic",
    "rovers_fond",
    "social-planning",
    "sokoban-sequential-optimal",
    "spiky-tireworld",
    "storage",
    "tireworld",
    "tireworld-truck",
    "triangle-tireworld",
    "zenotravel",
    "hello-world-functions",
]

PROBLEM_FILES = list(
    itertools.chain(
        *[
            (FIXTURES_PDDL_FILES / domain_name).rglob("p*.pddl")
            for domain_name in DOMAIN_NAMES
        ]
    )
)

PLAN_FILES = list(
    itertools.chain(
        *[
            (FIXTURES_PDDL_FILES / domain_name).rglob("p*.plan")
            for domain_name in DOMAIN_NAMES
        ]
    )
)


@pytest.fixture(scope="session")
def domain_parser():
    """Get the PDDL domain parser."""
    return DomainParser()


@pytest.fixture(scope="session")
def problem_parser():
    """Get the PDDL problem parser."""
    return ProblemParser()


@pytest.fixture(scope="session")
def markdown_parser():
    """Get the Mistune Markdown parser."""
    return mistune.create_markdown(renderer=mistune.AstRenderer())


#################################################
# Import PDDL fixtures
from tests.fixtures.code_objects.blocksworld_fond import (  # noqa: E402, F401
    blocksworld_fond_01,
    blocksworld_fond_domain,
)
from tests.fixtures.code_objects.blocksworld_ipc08 import (  # noqa: E402, F401
    blocksworld_domain,
    blocksworld_problem_01,
)
from tests.fixtures.code_objects.triangle_tireworld import (  # noqa: E402, F401
    triangle_tireworld_domain,
    triangle_tireworld_problem_01,
)

pddl_objects_domains = [
    lf("blocksworld_domain"),  # type: ignore
    lf("triangle_tireworld_domain"),  # type: ignore
    lf("blocksworld_fond_domain"),  # type: ignore
]
pddl_objects_problems = [
    lf("blocksworld_fond_01"),  # type: ignore
    lf("blocksworld_problem_01"),  # type: ignore
    lf("triangle_tireworld_problem_01"),  # type: ignore
]

#################################################


# A set of symbols that can be matched as names but they are keywords
# this is a subset of all symbols
TEXT_SYMBOLS = {
    Symbols.AND.value,
    Symbols.DEFINE.value,
    Symbols.DOMAIN.value,
    Symbols.EITHER.value,
    Symbols.EXISTS.value,
    Symbols.FORALL.value,
    Symbols.NOT.value,
    Symbols.OBJECT.value,
    Symbols.ONEOF.value,
    Symbols.OR.value,
    Symbols.PROBLEM.value,
    Symbols.WHEN.value,
}



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains parametrized tests for both domains and problems parsers."""

from pathlib import Path

import pytest
from pytest_lazy_fixtures import lf as lazy_fixture  # type: ignore  # noqa

from pddl.core import Domain, Problem
from tests.conftest import (
    BLOCKSWORLD_FILES,
    BLOCKSWORLD_FOND_FILES,
    PROBLEM_FILES,
    TRIANGLE_FILES,
)


@pytest.mark.parametrize("pddl_file", PROBLEM_FILES)
def test_problem_parser(domain_parser, problem_parser, pddl_file: Path):
    """Test problem parsing and validation works for all the fixtures."""
    problem = problem_parser(pddl_file.read_text())
    domain_file = pddl_file.parent / "domain.pddl"
    domain = domain_parser(domain_file.read_text())
    problem.check(domain)


@pytest.mark.parametrize(
    "pddl_file,expected_problem",
    [
        (
            BLOCKSWORLD_FILES / "p01.pddl",
            lazy_fixture("blocksworld_problem_01"),  # type: ignore
        ),
        (
            BLOCKSWORLD_FOND_FILES / "p01.pddl",
            lazy_fixture("blocksworld_fond_01"),  # type: ignore
        ),
    ],
)
def test_check_problem_parser_output(problem_parser, pddl_file: Path, expected_problem):
    """Test problem parsing."""
    actual_problem = problem_parser(pddl_file.read_text())

    assert isinstance(actual_problem, Problem)
    assert actual_problem == expected_problem


@pytest.mark.parametrize(
    "pddl_file,expected_domain",
    [
        (
            BLOCKSWORLD_FILES / "domain.pddl",
            lazy_fixture("blocksworld_domain"),  # type: ignore
        ),
        (
            TRIANGLE_FILES / "domain.pddl",
            lazy_fixture("triangle_tireworld_domain"),  # type: ignore
        ),
        (
            BLOCKSWORLD_FOND_FILES / "domain.pddl",
            lazy_fixture("blocksworld_fond_domain"),  # type: ignore
        ),
    ],
)
def test_check_domain_parser_output(domain_parser, pddl_file: Path, expected_domain):
    """Test domain parsing."""
    actual_domain = domain_parser(pddl_file.read_text())

    assert isinstance(actual_domain, Domain)
    assert actual_domain == expected_domain



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Package metadata declaration."""

__title__ = "pddl"
__description__ = "PDDL parser"
__url__ = "https://github.com/AI-Planning/pddl.git"
__version__ = "0.4.7"
__author__ = "Marco Favorito, Francesco Fuggitti, Christian Muise"
__author_email__ = (
    "marco.favorito@gmail.com, francesco.fuggitti@gmail.com, christian.muise@queensu.ca"
)
__license__ = "MIT License"
__copyright__ = "2021-2025 WhiteMech"



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#
"""This module contains tests for a PDDL action."""

from pddl.action import Action
from pddl.logic import Predicate, Variable, variables
from pddl.logic.base import ExistsCondition, ForallCondition, Imply, OneOf


class TestActionEmpty:
    """Test the empty action."""

    def setup_method(self):
        """Set up the tests."""
        self.action = Action("empty_action", [])

    def test_name(self):
        """Test the name getter."""
        assert self.action.name == "empty_action"

    def test_parameters(self):
        """Test the parameters getter."""
        assert self.action.parameters == ()

    def test_precondition(self):
        """Test the precondition getter."""
        assert self.action.precondition is None

    def test_effect(self):
        """Test the effect getter."""
        assert self.action.effect is None


def test_build_simple_action():
    """Test a simple PDDL action."""
    x, y, z = variables("x y z", types=["type1"])
    p = Predicate("p", x, y, z)
    q = Predicate("q", x, y, z)
    action = Action(
        "simple_action",
        parameters=(x, y, z),
        precondition=p & ~q,
        effect=~p & q,
    )
    assert action


def test_build_action_1():
    """Test PDDL action 1."""
    x, y = variables("x y", types=["type1"])
    z = Variable("z", type_tags={"type2"})
    p = Predicate("p", x, y)
    q = Predicate("q", z)
    action = Action(
        "action_1",
        parameters=(x, y, z),
        precondition=p,
        effect=OneOf(p & ~q, p & q),
    )
    assert action


def test_build_action_2():
    """Test PDDL action 2."""
    x, y = variables("x y", types=["type1"])
    z, w = variables("z w", types=["type2"])
    p = Predicate("p", x, y)
    q = Predicate("q", z)
    r = Predicate("r", w)
    action = Action(
        "action_1",
        parameters=(x, y, z, w),
        precondition=Imply(p & q, r),
        effect=OneOf(p & ~q, p & q),
    )
    assert action


def test_build_action_quantified_conditions():
    """Test PDDL action with quantified conditions."""
    # set up variables and constants
    x, y, z, z2 = variables("x y z z2", types=["type_1"])

    # define predicates
    p1 = Predicate("p1", x, y, z)
    p2 = Predicate("p2", x, y)

    # define actions
    c1 = ExistsCondition(cond=~p2(y, z2), variables=[z2])
    c2 = ForallCondition(cond=p2(y, z2) & p1(x, y, z), variables=[z2])
    action = Action(
        "action-1", parameters=[x, y, z], precondition=c1 & c2, effect=p2(y, z)
    )

    assert action



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module defines validation functions for PDDL data structures."""

import functools
from collections.abc import Iterable
from typing import AbstractSet, Collection, Dict, Optional, Set, Tuple, cast

from pddl.action import Action
from pddl.custom_types import _check_not_a_keyword  # noqa: F401
from pddl.custom_types import name as name_type
from pddl.custom_types import namelike, to_names, to_types  # noqa: F401
from pddl.exceptions import PDDLValidationError
from pddl.helpers.base import check, ensure, ensure_set, find_cycle
from pddl.logic import Predicate
from pddl.logic.base import And, BinaryOp, QuantifiedCondition, UnaryOp
from pddl.logic.effects import Forall, When
from pddl.logic.functions import (
    BinaryFunction,
    FunctionExpression,
    NumericFunction,
    NumericValue,
)
from pddl.logic.predicates import DerivedPredicate, EqualTo
from pddl.logic.terms import Term
from pddl.parser.symbols import Symbols
from pddl.requirements import Requirements, _extend_domain_requirements


def validate(condition: bool, message: str = "") -> None:
    """
    Validate a condition regarding PDDL.

    If the condition is not satisfied, a PDDLValidationError is raised.
    """
    check(condition, message, PDDLValidationError)


def _find_inconsistencies_in_typed_terms(
    terms: Optional[Collection[Term]], all_types: Set[name_type]
) -> Optional[Tuple[Term, name_type]]:
    """
    Check that the terms in input all have legal types according to the list of available types.

    :param terms: the terms to check
    :param all_types: all available types
    :return: the type tag that raised the error, None otherwise
    """
    if terms is None:
        return None
    for term in terms:
        for type_tag in sorted(term.type_tags):
            if type_tag is not None and type_tag not in all_types:
                return term, type_tag
    return None


def _check_types_in_has_terms_objects(
    has_terms_objects: Optional[Collection[Predicate]],
    all_types: Set[name_type],
) -> None:
    """Check that the terms in the set of predicates all have legal types."""
    if has_terms_objects is None:
        return

    for has_terms in has_terms_objects:
        check_result = _find_inconsistencies_in_typed_terms(has_terms.terms, all_types)
        if check_result is not None:
            term, type_tag = check_result
            raise PDDLValidationError(
                f"type {repr(type_tag)} of term {repr(term)} in atomic expression "
                f"{repr(has_terms)} is not in available types {all_types}"
            )


class Types:
    """A class for representing and managing the types available in a PDDL Domain."""

    def __init__(
        self,
        types: Optional[Dict[namelike, Optional[namelike]]] = None,
        requirements: Optional[AbstractSet[Requirements]] = None,
        skip_checks: bool = False,
    ) -> None:
        """Initialize the Types object."""
        self._types = to_types(ensure(types, dict()))

        self._all_types = self._get_all_types()

        if not skip_checks:
            self._check_types_dictionary(self._types, ensure_set(requirements))

    @property
    def raw(self) -> Dict[name_type, Optional[name_type]]:
        """Get the raw types dictionary."""
        return self._types

    @property
    def all_types(self) -> Set[name_type]:
        """Get all available types."""
        return self._all_types

    def _get_all_types(self) -> Set[name_type]:
        """Get all types supported by the domain."""
        if self._types is None:
            return set()
        result = set(self._types.keys()) | set(self._types.values())
        result.discard(None)
        return cast(Set[name_type], result)

    @classmethod
    def _check_types_dictionary(
        cls,
        type_dict: Dict[name_type, Optional[name_type]],
        requirements: AbstractSet[Requirements],
    ) -> None:
        """
        Check the consistency of the types dictionary.

        1) Empty types dictionary is correct by definition:
        >>> Types._check_types_dictionary({}, set())

        2) There are supertypes, but :typing requirement not specified
        >>> a, b, c = to_names(["a", "b", "c"])
        >>> Types._check_types_dictionary({a: b, b: c}, set())
        Traceback (most recent call last):
        ...
        pddl.exceptions.PDDLValidationError: typing requirement is not specified, but types are used: 'b', 'c'

        3) The `object` type cannot be a subtype:
        >>> a = name_type("a")
        >>> Types._check_types_dictionary({name_type("object"): a}, {Requirements.TYPING})
        Traceback (most recent call last):
        ...
        pddl.exceptions.PDDLValidationError: object must not have supertypes, but got 'object' is a subtype of 'a'

        4) If cycles in the type hierarchy graph are present, an error is raised:
        >>> a, b, c = to_names(["a", "b", "c"])
        >>> Types._check_types_dictionary({a: b, b: c, c: a}, {Requirements.TYPING})
        Traceback (most recent call last):
        ...
        pddl.exceptions.PDDLValidationError: cycle detected in the type hierarchy: a -> b -> c

        :param type_dict: the types dictionary
        """
        if len(type_dict) == 0:
            return

        # check typing requirement
        supertypes = {t for t in type_dict.values() if t is not None}
        if len(supertypes) > 0 and Requirements.TYPING not in requirements:
            raise PDDLValidationError(
                "typing requirement is not specified, but types are used: '"
                + "', '".join(map(str, sorted(supertypes)))
                + "'"
            )

        # check `object` type
        object_name = name_type(Symbols.OBJECT.value)
        if object_name in type_dict and type_dict[object_name] is not None:
            object_supertype = type_dict[object_name]
            raise PDDLValidationError(
                f"object must not have supertypes, but got 'object' is a subtype of '{object_supertype}'"
            )

        # check cycles
        # need to convert type_dict to a dict of sets, because find_cycle() expects a dict of sets
        cycle = find_cycle(
            {
                key: {value} if value is not None else set()
                for key, value in type_dict.items()
            }
        )  # type: ignore
        if cycle is not None:
            raise PDDLValidationError(
                "cycle detected in the type hierarchy: " + " -> ".join(cycle)
            )


class TypeChecker:
    """Implementation of a type checker for domains and problems."""

    def __init__(
        self, types: Types, requirements: Optional[AbstractSet[Requirements]] = None
    ) -> None:
        """Initialize the type checker."""
        self._types = types
        self._requirements = ensure_set(requirements)

    @property
    def has_typing(self) -> bool:
        """Check if the typing requirement is specified."""
        self._extended_requirements = _extend_domain_requirements(self._requirements)
        return Requirements.TYPING in self._extended_requirements

    def _check_typing_requirement(self, type_tags: Collection[name_type]) -> None:
        """Check that the typing requirement is specified."""
        if not self.has_typing and len(type_tags) > 0:
            raise PDDLValidationError(
                f"typing requirement is not specified, but the following types were used: {type_tags}"
            )

    def _check_types_are_available(
        self, type_tags: Collection[name_type], what: str
    ) -> None:
        """Check that the types are available in the domain."""
        if not self._types.all_types.issuperset(type_tags):
            raise PDDLValidationError(
                f"types {sorted(type_tags)} of {what} are not in available types {self._types.all_types}"
            )

    @functools.singledispatchmethod  # type: ignore
    def check_type(self, obj: object):
        """Check types annotations of PDDL data structures."""
        raise NotImplementedError(f"cannot check PDDL types of {type(obj)}")

    @check_type.register
    def _(self, objects: Iterable) -> None:
        """Check the types of collections of objects."""
        for obj in objects:
            self.check_type(obj)

    @check_type.register
    def _(self, term: Term) -> None:
        """Check types annotations of a PDDL term."""
        self._check_typing_requirement(term.type_tags)
        self._check_types_are_available(term.type_tags, f"term {repr(term)}")

    @check_type.register
    def _(self, predicate: Predicate) -> None:
        """Check types annotations of a PDDL predicate."""
        self.check_type(predicate.terms)

    @check_type.register
    def _(self, function: NumericFunction) -> None:
        """Check types annotations of a PDDL numeric function."""
        self.check_type(function.terms)

    @check_type.register
    def _(self, _: NumericValue) -> None:
        """Check types annotations of a PDDL numeric value operator."""
        return None

    @check_type.register
    def _(self, binary_function: BinaryFunction) -> None:
        """Check types annotations of a PDDL numeric binary operator."""
        self.check_type(binary_function.operands)

    @check_type.register
    def _(self, equal_to: EqualTo) -> None:
        """Check types annotations of a PDDL equal-to atomic formula."""
        self.check_type(equal_to.left)
        self.check_type(equal_to.right)

    @check_type.register
    def _(self, derived_predicate: DerivedPredicate) -> None:
        """Check types annotations of a PDDL derived predicate."""
        self.check_type(derived_predicate.predicate)
        self.check_type(derived_predicate.condition)

    @check_type.register
    def _(self, formula: UnaryOp) -> None:
        """Check types annotations of a PDDL unary operator."""
        self.check_type(formula.argument)

    @check_type.register
    def _(self, formula: BinaryOp) -> None:
        """Check types annotations of a PDDL binary operator."""
        self.check_type(formula.operands)

    @check_type.register
    def _(self, formula: QuantifiedCondition) -> None:
        """Check types annotations of a PDDL quantified condition."""
        self.check_type(formula.variables)
        self.check_type(formula.condition)

    @check_type.register
    def _(self, effect: And) -> None:
        """Check types annotations of a PDDL and-effect."""
        self.check_type(effect.operands)

    @check_type.register
    def _(self, effect: When) -> None:
        """Check types annotations of a PDDL when-effect."""
        self.check_type(effect.condition)
        self.check_type(effect.effect)

    @check_type.register
    def _(self, effect: Forall) -> None:
        """Check types annotations of a PDDL forall-effect."""
        self.check_type(effect.variables)
        self.check_type(effect.effect)

    @check_type.register
    def _(self, action: Action) -> None:
        """Check types annotations of a PDDL term."""
        self.check_type(action.parameters)
        self.check_type(action.precondition)
        self.check_type(action.effect)


class Functions:
    """A class for representing and managing the numeric functions available in a PDDL Domain."""

    def __init__(
        self,
        functions: Optional[Collection[FunctionExpression]] = None,
        requirements: Optional[AbstractSet[Requirements]] = None,
        skip_checks: bool = False,
    ) -> None:
        """Initialize the Functions object."""
        self._functions = ensure(functions, dict())
        self._all_function = self._get_all_functions()

        if not skip_checks:
            self._check_total_cost(self._functions, ensure_set(requirements))

    @property
    def raw(self) -> Dict[NumericFunction, Optional[name_type]]:
        """Get the raw functions' dictionary."""
        return self._functions

    @property
    def all_functions(self) -> Set[NumericFunction]:
        """Get all available functions."""
        return self._all_function

    def _get_all_functions(self) -> Set[NumericFunction]:
        """Get all function supported by the domain."""
        if self._functions is None:
            return set()
        result = set(self._functions.keys()) | set(self._functions.values())
        result.discard(None)
        return cast(Set[NumericFunction], result)

    @classmethod
    def _check_total_cost(
        cls,
        function_dict: Dict[NumericFunction, Optional[name_type]],
        requirements: AbstractSet[Requirements],
    ) -> None:
        """Check consistency of total-cost function with the requirements."""
        if bool(function_dict):
            if any(f.name == Symbols.TOTAL_COST.value for f in {*function_dict}):
                validate(
                    Requirements.ACTION_COSTS in requirements,
                    f"action costs requirement is not specified, but the {Symbols.TOTAL_COST.value} "
                    f"function is specified.",
                )
                if any(
                    isinstance(f, FunctionExpression)
                    and not f.name == Symbols.TOTAL_COST.value
                    for f in function_dict.keys()
                ):
                    validate(
                        Requirements.NUMERIC_FLUENTS
                        in _extend_domain_requirements(requirements),
                        "numeric-fluents requirement is not specified, but numeric fluents are specified.",
                    )
            else:
                validate(
                    Requirements.NUMERIC_FLUENTS
                    in _extend_domain_requirements(requirements),
                    "numeric-fluents requirement is not specified, but numeric fluents are specified.",
                )



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains tests for PDDL predicates."""

from pddl.core import Predicate
from pddl.logic.base import And
from pddl.logic.helpers import constants, variables
from pddl.logic.predicates import DerivedPredicate, EqualTo


class TestPredicateSimpleInitialisation:
    """Test simple predicate initialisation."""

    def setup_method(self):
        """Set up the tests."""
        self.a, self.b = variables("a b")
        self.c = constants("c")[0]
        self.predicate = Predicate("P", self.a, self.b)
        self.predicate2 = Predicate("P2", self.a, self.c)

    def test_name(self):
        """Test name getter."""
        assert self.predicate.name == "P"

    def test_variables(self):
        """Test terms getter."""
        assert self.predicate.terms == (self.a, self.b)

    def test_arity(self):
        """Test arity property."""
        assert self.predicate.arity == 2

    def test_to_equal(self):
        """Test to equal."""
        other = Predicate("P", self.a, self.b)
        assert self.predicate == other

    def test_to_str(self):
        """Test to string."""
        assert str(self.predicate) == f"({self.predicate.name} {self.a} {self.b})"
        assert str(self.predicate2) == f"({self.predicate2.name} {self.a} {self.c})"

    def test_to_repr(self):
        """Test to repr."""
        assert (
            repr(self.predicate)
            == f"Predicate({self.predicate.name}, {self.a}, {self.b})"
        )


class TestEqualToPredicate:
    """Test the equal to predicate."""

    def setup_method(self):
        """Set up the tests."""
        self.left, self.right = variables("l r")
        self.equal_to = EqualTo(self.left, self.right)

    def test_left(self):
        """Test left getter."""
        assert self.equal_to.left == self.left

    def test_right(self):
        """Test right getter."""
        assert self.equal_to.right == self.right

    def test_to_equal(self):
        """Test to equal."""
        other = EqualTo(self.left, self.right)
        assert self.equal_to == other

    def test_to_str(self):
        """Test to string."""
        assert str(self.equal_to) == f"(= {str(self.left)} {str(self.right)})"

    def test_to_repr(self):
        """Test to repr."""
        assert repr(self.equal_to) == f"EqualTo({repr(self.left)}, {repr(self.right)})"


class TestDerivedPredicate:
    """Test the derived predicate."""

    def setup_method(self):
        """Set up the tests."""
        self.a, self.b = variables("a b")
        self.predicate = Predicate("dp", self.a, self.b)
        self.condition = And(self.a, self.b)
        self.derived = DerivedPredicate(self.predicate, self.condition)

    def test_predicate(self):
        """Test predicate getter."""
        assert self.derived.predicate == self.predicate

    def test_condition(self):
        """Test condition getter."""
        assert self.derived.condition == self.condition

    def test_to_equal(self):
        """Test to equal."""
        other = DerivedPredicate(self.predicate, self.condition)
        assert self.derived == other

    def test_to_str(self):
        """Test to string."""
        assert str(self.derived) == f"(:derived {self.predicate} {self.condition})"

    def test_to_repr(self):
        """Test to repr."""
        assert (
            repr(self.derived)
            == f"DerivedPredicate({repr(self.predicate)}, {repr(self.condition)})"
        )



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains tests for PDDL functions."""

import pytest

from pddl.logic.functions import Metric, NumericFunction, NumericValue
from pddl.logic.helpers import constants, variables
from pddl.parser.symbols import Symbols


class TestNumericFunction:
    """Test simple numeric function."""

    def setup_method(self):
        """Set up the tests."""
        self.a, self.b = variables("a b", types=["type1"])
        self.c = constants("c")[0]
        self.function = NumericFunction("func", self.a, self.b)
        self.function2 = NumericFunction("func2", self.a, self.c)

    def test_name(self):
        """Test name getter."""
        assert self.function.name == "func"

    def test_variables(self):
        """Test terms getter."""
        assert self.function.terms == (self.a, self.b)

    def test_arity(self):
        """Test arity property."""
        assert self.function.arity == 2

    def test_to_equal(self):
        """Test to equal."""
        other = NumericFunction("func", self.a, self.b)
        assert self.function == other

    def test_to_str(self):
        """Test to string."""
        assert str(self.function) == f"({self.function.name} {self.a} {self.b})"
        assert str(self.function2) == f"({self.function2.name} {self.a} {self.c})"

    def test_to_repr(self):
        """Test to repr."""
        assert (
            repr(self.function)
            == f"NumericFunction({self.function.name}, {self.a}, {self.b})"
        )


class TestTotalCost:
    """Test total cost function."""

    def setup_method(self):
        """Set up the tests."""
        self.total_cost = NumericFunction(Symbols.TOTAL_COST.value)

    def test_name(self):
        """Test name getter."""
        assert self.total_cost.name == "total-cost"


class TestMetric:
    """Test metric."""

    def setup_method(self):
        """Set up the tests."""
        self.a, self.b = variables("a b")
        self.function = NumericFunction("func", self.a, self.b)
        self.maximize_metric = Metric(self.function, Metric.MAXIMIZE)
        self.minimize_metric = Metric(self.function, Metric.MINIMIZE)

    def test_function_maximize(self):
        """Test function getter for maximize metric."""
        assert self.maximize_metric.expression == self.function

    def test_function_minimize(self):
        """Test function getter for minimize metric."""
        assert self.minimize_metric.expression == self.function

    def test_optimization_maximize(self):
        """Test optimization getter for maximize metric."""
        assert self.maximize_metric.optimization == Metric.MAXIMIZE

    def test_optimization_minimize(self):
        """Test optimization getter for minimize metric."""
        assert self.minimize_metric.optimization == Metric.MINIMIZE

    def test_wrong_optimization(self):
        """Test wrong optimization."""
        with pytest.raises(
            AssertionError,
            match="Optimization metric not recognized.",
        ):
            Metric(self.function, "other")

    def test_to_equal(self):
        """Test to equal."""
        other = Metric(NumericFunction("func", self.a, self.b), Metric.MINIMIZE)
        assert self.minimize_metric == other

    def test_to_str(self):
        """Test to string."""
        assert (
            str(self.maximize_metric)
            == f"{self.maximize_metric.optimization} {self.maximize_metric.expression}"
        )

    def test_to_repr(self):
        """Test to repr."""
        assert (
            repr(self.maximize_metric)
            == f"Metric({self.maximize_metric.expression}, {self.maximize_metric.optimization})"
        )


class TestNumericValue:
    """Test the numeric value."""

    def setup_method(self):
        """Set up the tests."""
        self.numeric_value = NumericValue(3)

    def test_value(self):
        """Test the name getter."""
        assert self.numeric_value.value == 3



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Test the cache hash class decorator."""

import pickle  # nosec

from pddl.helpers.cache_hash import cache_hash


@cache_hash
class MyHashable:
    """A test class to test 'Hashable' metaclass."""

    def __init__(self):
        """Initialize."""
        super().__init__()
        self.a = "a"
        self.b = "b"

    def __hash__(self):
        """Compute the hash."""
        return hash((self.a, self.b))


def test_hashable():
    """Test the hashable class."""
    obj = MyHashable()

    assert not hasattr(obj, "__hash")

    h1 = hash(obj)
    h2 = hash(obj)
    assert h1 == h2

    assert hasattr(obj, "__hash")
    assert obj.__hash == h1 == h2

    dumped_obj = pickle.dumps(obj)  # nosec
    actual_obj = pickle.loads(dumped_obj)  # nosec
    assert not hasattr(actual_obj, "__hash")



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This class implements PDDL predicates."""

import functools
from typing import Dict, Mapping, Sequence, Set

from pddl.custom_types import name, namelike, parse_name
from pddl.helpers.base import assert_, check
from pddl.helpers.cache_hash import cache_hash
from pddl.logic.base import Atomic, Formula
from pddl.logic.terms import Term, Variable, _print_tag_set
from pddl.parser.symbols import Symbols


def _check_terms_consistency(terms: Sequence[Term]):
    """
    Check that the term sequence have consistent type tags.

    In particular, terms with the same name must have the same type tags.
    """
    seen: Dict[name, Set[name]] = {}
    for term in terms:
        if term.name not in seen:
            seen[term.name] = set(term.type_tags)
        else:
            check(
                seen[term.name] == set(term.type_tags),
                f"Term {term} has inconsistent type tags: "
                f"previous type tags {_print_tag_set(seen[term.name])}, new type tags {_print_tag_set(term.type_tags)}",
                exception_cls=ValueError,
            )


@cache_hash
@functools.total_ordering
class Predicate(Atomic):
    """A class for a Predicate in PDDL."""

    def __init__(self, predicate_name: namelike, *terms: Term):
        """Initialize the predicate."""
        self._name = parse_name(predicate_name)
        self._terms = tuple(terms)
        _check_terms_consistency(self._terms)

    @property
    def name(self) -> str:
        """Get the name."""
        return self._name

    @property
    def terms(self) -> Sequence[Term]:
        """Get the terms."""
        return self._terms

    @property
    def arity(self) -> int:
        """Get the arity of the predicate."""
        return len(self.terms)

    def instantiate(self, mapping: Mapping[Variable, Term]) -> "Predicate":
        """Instantiate the formula with a mapping from variables to terms."""
        instantiated_terms = []
        for term in self.terms:
            if isinstance(term, Variable) and term in mapping:
                # Instantiate the variable
                instantiated_terms.append(mapping[term])
            else:
                # The predicate is already partially instantiated or mapping is a partial instantiation
                instantiated_terms.append(term)
        return Predicate(self.name, *instantiated_terms)

    def __call__(self, *terms: Term):
        """Replace terms."""
        assert_(len(terms) == self.arity, "Number of terms not correct.")
        return Predicate(self.name, *terms)

    def __str__(self) -> str:
        """Get the string."""
        if self.arity == 0:
            return f"({self.name})"
        else:
            return f"({self.name} {' '.join(map(str, self.terms))})"

    def __repr__(self) -> str:
        """Get an unambiguous string representation."""
        return f"{type(self).__name__}({self.name}, {', '.join(map(str, self.terms))})"

    def __eq__(self, other):
        """Override equal operator."""
        return (
            isinstance(other, Predicate)
            and self.name == other.name
            and self.terms == other.terms
        )

    def __hash__(self):
        """Get the has of a Predicate."""
        return hash((self.name, self.arity, self.terms))

    def __lt__(self, other):
        """Compare with another object."""
        if isinstance(other, Predicate):
            return (self.name, self.terms) < (other.name, other.terms)
        return super().__lt__(other)


class EqualTo(Atomic):
    """Equality predicate."""

    def __init__(self, left: Term, right: Term):
        """
        Initialize the equality predicate.

        :param left: the left term.
        :param right: the right term.
        """
        self._left = left
        self._right = right
        _check_terms_consistency([self._left, self._right])

    @property
    def left(self) -> Term:
        """Get the left operand."""
        return self._left

    @property
    def right(self) -> Term:
        """Get the right operand."""
        return self._right

    def instantiate(self, mapping: Mapping[Variable, Term]) -> "EqualTo":
        """Instantiate the formula with a mapping from variables to terms."""
        instantiated_left = self._left
        instantiated_right = self._right
        if isinstance(self._left, Variable) and self._left in mapping:
            instantiated_left = mapping[self._left]
        if isinstance(self._right, Variable) and self._right in mapping:
            instantiated_right = mapping[self._right]
        return EqualTo(instantiated_left, instantiated_right)

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return (
            isinstance(other, EqualTo)
            and self.left == other.left
            and self.right == other.right
        )

    def __hash__(self) -> int:
        """Get the hash."""
        return hash((EqualTo, self.left, self.right))

    def __str__(self) -> str:
        """Get the string representation."""
        return f"({Symbols.EQUAL.value} {self.left} {self.right})"

    def __repr__(self) -> str:
        """Get the string representation."""
        return f"{type(self).__name__}({repr(self.left)}, {repr(self.right)})"


@cache_hash
@functools.total_ordering
class DerivedPredicate(Atomic):
    """A class for a Derived Predicates in PDDL."""

    def __init__(self, predicate: Predicate, condition: Formula) -> None:
        """
        Initialize the derived predicate.

        :param predicate: the predicate
        :param condition: the condition
        """
        self._predicate = predicate
        self._condition = condition

    @property
    def predicate(self) -> Predicate:
        """Get the predicate."""
        return self._predicate

    @property
    def condition(self) -> Formula:
        """Get the condition."""
        return self._condition

    def instantiate(self, mapping: Mapping[Variable, Term]) -> "DerivedPredicate":
        """Instantiate the formula with a mapping from variables to terms."""
        return DerivedPredicate(
            self.predicate.instantiate(mapping),
            self.condition.instantiate(mapping),
        )

    def __hash__(self) -> int:
        """Get the hash."""
        return hash((DerivedPredicate, self.predicate, self.condition))

    def __str__(self) -> str:
        """Get the string representation."""
        return f"({Symbols.DERIVED.value} {self.predicate} {self.condition})"

    def __repr__(self) -> str:
        """Get the string representation."""
        return f"{type(self).__name__}({repr(self.predicate)}, {repr(self.condition)})"

    def __eq__(self, other):
        """Override equal operator."""
        return (
            isinstance(other, DerivedPredicate)
            and self.predicate == other.predicate
            and self.condition == other.condition
        )

    def __lt__(self, other):
        """Compare with another object."""
        if isinstance(other, DerivedPredicate):
            return (self.predicate, self._predicate) < (
                other.predicate,
                other.condition,
            )
        return super().__lt__(other)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains a utility function to propagate type tags of variables correctly along the formula."""

import copy
from functools import singledispatch
from typing import FrozenSet, Mapping, Optional

from pddl.custom_types import namelike
from pddl.logic.base import BinaryOp, Formula, QuantifiedCondition, UnaryOp
from pddl.logic.functions import BinaryFunction, NumericFunction, NumericValue
from pddl.logic.predicates import EqualTo as EqualToPredicate
from pddl.logic.predicates import Predicate
from pddl.logic.terms import Term, Variable


def update_type_tags(
    formula: Formula,
    var_to_types: Mapping[str, Optional[FrozenSet[namelike]]],
):
    """Update type tags of variables in a formula."""
    return _update_type_tags(formula, var_to_types)


@singledispatch
def _update_type_tags(
    formula: Formula,
    var_to_types: Mapping[str, Optional[FrozenSet[namelike]]],
):
    raise NotImplementedError(f"Formula of type {type(formula)} cannot be processed.")


@_update_type_tags.register
def _update_type_tags_binary_op(
    formula: BinaryOp, var_to_types: Mapping[str, Optional[FrozenSet[namelike]]]
):
    new_operands = []
    for operand in formula.operands:
        new_operand = _update_type_tags(operand, var_to_types)
        new_operands.append(new_operand)
    return formula.__class__(*new_operands)


@_update_type_tags.register
def _update_type_tags_unary_op(
    formula: UnaryOp, var_to_types: Mapping[str, Optional[FrozenSet[namelike]]]
):
    new_operand = _update_type_tags(formula.argument, var_to_types)
    return formula.__class__(new_operand)


@_update_type_tags.register
def _update_type_tags_quantified_condition(
    formula: QuantifiedCondition,
    var_to_types: Mapping[str, Optional[FrozenSet[namelike]]],
):
    new_var_to_types = dict(copy.copy(var_to_types))
    for var in formula.variables:
        new_var_to_types[var.name] = frozenset(var.type_tags)
    new_operand = _update_type_tags(formula.condition, new_var_to_types)
    return formula.__class__(new_operand, list(formula.variables))


@_update_type_tags.register
def _update_type_tags_equal_to_predicate(
    formula: EqualToPredicate,
    var_to_types: Mapping[str, Optional[FrozenSet[namelike]]],
):
    left_operand = _replace_term_with_type_tags_if_var(formula.left, var_to_types)
    right_operand = _replace_term_with_type_tags_if_var(formula.right, var_to_types)
    return EqualToPredicate(left_operand, right_operand)


@_update_type_tags.register
def _update_type_tags_predicate(
    formula: Predicate, var_to_types: Mapping[str, Optional[FrozenSet[namelike]]]
):
    new_terms = []
    for term in formula.terms:
        new_terms.append(_replace_term_with_type_tags_if_var(term, var_to_types))
    return Predicate(formula.name, *new_terms)


@_update_type_tags.register
def _update_type_tags_numeric_function(
    formula: NumericFunction, var_to_types: Mapping[str, Optional[FrozenSet[namelike]]]
):
    new_terms = []
    for term in formula.terms:
        new_terms.append(_replace_term_with_type_tags_if_var(term, var_to_types))
    return NumericFunction(formula.name, *new_terms)


@_update_type_tags.register
def _update_type_tags_numeric_value(
    formula: NumericValue, var_to_types: Mapping[str, Optional[FrozenSet[namelike]]]
):
    return copy.copy(formula)


@_update_type_tags.register
def _update_type_tags_binary_function(
    formula: BinaryFunction, var_to_types: Mapping[str, Optional[FrozenSet[namelike]]]
):
    new_operands = []
    for operand in formula.operands:
        new_operand = _update_type_tags(operand, var_to_types)
        new_operands.append(new_operand)
    return formula.__class__(*new_operands)


def _replace_term_with_type_tags_if_var(
    term: Term, var_to_types: Mapping[str, Optional[FrozenSet[namelike]]]
) -> Term:
    if isinstance(term, Variable) and term.name in var_to_types:
        return term.with_type_tags(var_to_types[term.name])
    return term



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Test the base helpers."""

from pddl.helpers.base import find_cycle


def test_find_cycle_empty_graph() -> None:
    """Test 'find_cycle' function, empty graph."""
    assert find_cycle({}) is None


def test_find_cycle_negative_case() -> None:
    """Test 'find_cycle' function, negative case."""
    a, b, c, d = "a b c d".split()
    assert find_cycle({a: {b}, b: {c}, d: {c}}) is None


def test_find_cycle_positive_case() -> None:
    """Test 'find_cycle' function, positive case."""
    a, b, c, d = "a b c d".split()
    assert find_cycle({a: {b}, b: {c}, d: {c}, c: {a}}) == ["a", "b", "c"]



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Test pddl.logic.terms module."""

import pytest

from pddl.logic.terms import Constant, Variable


def test_no_duplicated_type_tags() -> None:
    """Test that no duplicated type tags are allowed."""
    with pytest.raises(
        ValueError, match=r"duplicate element in collection \['b', 'b'\]: 'b'"
    ):
        Variable("a", ["b", "b"])


def test_variable_ordering() -> None:
    """Test variable ordering."""
    v1 = Variable("x", ["type1"])
    v2 = Variable("y", ["type3"])
    v3 = Variable("z", ["type2"])
    assert v1 < v2
    assert v1 < v3
    assert v2 < v3
    assert not (v1 > v2)
    assert not (v1 > v3)
    assert not (v2 > v3)


def test_constant_ordering() -> None:
    """Test constant ordering."""
    c1 = Constant("a", "type1")
    c2 = Constant("b", "type2")
    c3 = Constant("c", "type1")
    assert c1 < c2
    assert c1 < c3
    assert c2 < c3
    assert not (c1 > c2)
    assert not (c1 > c3)
    assert not (c2 > c3)


def test_term_ordering() -> None:
    """Test term ordering between variables and constants."""
    v1 = Variable("x", ["type1"])
    v2 = Variable("y", ["type3"])
    v3 = Variable("z", ["type2"])
    c1 = Constant("a", "type1")
    c2 = Constant("b", "type2")
    c3 = Constant("c", "type1")
    assert c1 < v1
    assert c2 < v2
    assert c3 < v3
    assert not (v1 < c1)
    assert not (v2 < c2)
    assert not (v3 < c3)
    array = [c1, c2, c3, v1, v2, v3]
    assert array == sorted(array)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module defines useful custom types."""

import re
from typing import AbstractSet, Collection, Dict, List, Optional, Union

from pddl.exceptions import PDDLValidationError
from pddl.helpers.base import RegexConstrainedString, ensure_set
from pddl.parser.symbols import ALL_SYMBOLS, Symbols


class name(RegexConstrainedString):
    """
    This type represents a 'name' in a PDDL file.

    It must match the following regex: "[A-Za-z][-_A-Za-z0-9]*".
    """

    REGEX = re.compile("[A-Za-z][-_A-Za-z0-9]*")


"""
Either a true name, or a string (potentially not a name!).
The purpose is to make the APIs more usable. The developer
should take care of converting the bare strings to names
for better consistency, when developing a library component;
this can be achieved thanks to 'name' constructor idempotency,
without explicitly caring of whether the arguments
are actually a 'name' or a 'str'.

Note, you should not use the raw constructor 'name' (unless you
know what you are doing), but the functino 'parse_name' below.
"""
namelike = Union[name, str]


def parse_name(s: str) -> name:
    """
    Parse a name from a string.

    It performs two validations:
    - the name is not a keyword;
    - the name is a valid name, i.e. it matches the name regular expression.

    Optionally, a set of keywords to ignore can be provided.

    :param s: the input string to be parsed
    :return: the parsed name
    """
    _check_not_a_keyword(s, "name", ignore=set())
    return name(s)


def parse_type(s: str) -> name:
    """
    Parse a type from a string.

    It performs two validations:
    - the type is not a keyword;
    - the type is a valid name, i.e. it matches the name regular expression.

    The type name 'object' is allowed.

    :param s: the input string to be parsed
    :return: the parsed type name
    """
    _check_not_a_keyword(s, "type", ignore={Symbols.OBJECT.value})
    return name(s)


def parse_function(s: str) -> name:
    """
    Parse a function name from a string.

    It performs two validations:
    - the function name is not a keyword;
    - the function name is a valid name, i.e. it matches the name regular expression.

    The function name 'total-cost' is allowed.

    :param s: the input string to be parsed
    :return: the parsed name
    """
    _check_not_a_keyword(s, "name", ignore={Symbols.TOTAL_COST.value})
    return name(s)


def to_names(names: Collection[namelike]) -> List[name]:
    """From name-like sequence to list of names."""
    return list(map(parse_name, names))


def to_type(names: Collection[namelike]) -> List[name]:
    """From name-like sequence to list of type names."""
    return list(map(parse_type, names))


def to_types(names: Dict[namelike, Optional[namelike]]) -> Dict[name, Optional[name]]:
    """From name-like dictionary to name dictionary."""
    return {
        parse_type(type_): parse_type(ancestor) if ancestor else None
        for type_, ancestor in names.items()
    }


def _is_a_keyword(word: str, ignore: Optional[AbstractSet[str]] = None) -> bool:
    """Check that the word is not a keyword."""
    ignore_set = ensure_set(ignore)
    # we remove the TOTAL_COST because it is not a keyword but a special function
    return word not in ignore_set and word in ALL_SYMBOLS


def _check_not_a_keyword(
    input_name: str,
    item_type: str,
    ignore: AbstractSet[str],
) -> None:
    """Check that the item name is not a keyword."""
    if _is_a_keyword(input_name, ignore=ignore):
        raise PDDLValidationError(
            f"invalid {item_type} '{input_name}': it is a keyword"
        )



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Helper functions."""

import re
from pathlib import Path
from typing import (
    AbstractSet,
    Any,
    Callable,
    Collection,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Type,
)


def _get_current_path() -> Path:
    """Get the path to the file where the function is called."""
    import inspect
    import os

    return Path(os.path.dirname(inspect.getfile(inspect.currentframe()))).parent  # type: ignore


def assert_(condition: bool, message: str = "") -> None:
    """
    User-defined assert.

    This function is useful to avoid the use of the built-in assert statement, which is removed
        when the code is compiled in optimized mode. For more information, see
        https://bandit.readthedocs.io/en/1.7.5/plugins/b101_assert_used.html
    """
    check(condition, message=message, exception_cls=AssertionError)


def check(
    condition: bool, message: str = "", exception_cls: Type[Exception] = AssertionError
) -> None:
    """Check a condition, and if false, raise exception."""
    if not condition:
        raise exception_cls(message)


def ensure(arg: Optional[Any], default: Any):
    """Ensure an object is not None, or return a default."""
    return arg if arg is not None else default


def ensure_set(arg: Optional[Collection], immutable: bool = True) -> AbstractSet:
    """
    Ensure the argument is a set.

    :param arg: the set, or None.
    :param immutable: whether the collection should be immutable.
    :return: the same set, or an empty set if the arg was None.
    """
    op = frozenset if immutable else set
    return op(arg) if arg is not None else op()


def check_no_duplicates(arg: Optional[Collection]) -> Optional[Collection]:
    """Check that the argument is a set."""
    if arg is None:
        return None
    if isinstance(arg, AbstractSet):
        return arg
    seen = set()
    for x in arg:
        if x in seen:
            raise ValueError(
                f"duplicate element in collection {list(map(str, arg))}: '{str(x)}'"
            )
        seen.add(x)
    return arg


def ensure_sequence(arg: Optional[Sequence], immutable: bool = True) -> Sequence:
    """
    Ensure the argument is a sequence.

    :param arg: the list, or None.
    :param immutable: whether the collection should be immutable.
    :return: the same list, or an empty list if the arg was None.
    """
    op: Type = tuple if immutable else list
    return op(arg) if arg is not None else op()


def safe_index(seq: Sequence, *args, **kwargs):
    """Find element, safe."""
    try:
        return seq.index(*args, **kwargs)
    except ValueError:
        return None


def safe_get(seq: Sequence, index: int, default=None):
    """Get element at index, safe."""
    return seq[index] if index < len(seq) else default


def find(seq: Sequence, condition: Callable[[Any], bool]) -> int:
    """
    Find the index of the first element that satisfies a condition.

    :param seq: the sequence.
    :param condition: the condition.
    :return: the index, or -1 if no element satisfies the condition.
    """
    return next((i for i, x in enumerate(seq) if condition(x)), -1)


def _typed_parameters(parameters) -> str:
    """Return a list of parameters along with types if available."""
    result = ""
    for i, p in enumerate(parameters):
        if i > 0:
            result += " "
        if p.type_tags:
            result += f"?{p.name} - {' '.join(map(str, p.type_tags))}"
        else:
            result += str(p)
    return result.strip()


class RegexConstrainedString(str):
    """
    A string that is constrained by a regex.

    The default behaviour is to match anything.
    Subclass this class and change the 'REGEX' class
    attribute to implement a different behaviour.
    """

    REGEX = re.compile(".*", flags=re.DOTALL)

    def __new__(cls, value, *args, **kwargs):
        """Instantiate a new object."""
        if type(value) is cls:
            return value
        else:
            inst = super(RegexConstrainedString, cls).__new__(cls, value)
            return inst

    def __init__(self, *_, **__):
        """Initialize a regex constrained string."""
        super().__init__()
        if not self.REGEX.fullmatch(self):
            self._handle_no_match()

    def _handle_no_match(self):
        raise ValueError(
            "Value '{data}' does not match the regular expression {regex}".format(
                data=self, regex=self.REGEX
            )
        )


def find_cycle(graph: Dict[str, Optional[AbstractSet[str]]]) -> Optional[Sequence[str]]:
    """Check whether a graph (represented as a dictionary-based adjacency list) has a cycle."""
    visited: Set = set()
    stack: List = []

    for node in graph:
        if node not in visited:
            stack.append((node, []))

            while stack:
                current, path = stack.pop()
                if current in path:
                    return path

                visited.add(current)
                neighbors = graph.get(current)
                if neighbors is not None:
                    for neighbor in neighbors:
                        stack.append((neighbor, path + [current]))

    return None



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains tests for the library custom types."""

import pytest

from pddl.custom_types import name, parse_name, parse_type
from pddl.exceptions import PDDLValidationError
from pddl.parser.symbols import Symbols
from tests.conftest import TEXT_SYMBOLS


def test_name_string():
    """Test the 'name' string subclass defined in pddl.types."""
    a = name("a")
    assert a == "a"


def test_name_constructor_twice():
    """Test that the name constructor is idempotent."""
    a0 = "a"
    a1 = name(a0)
    a2 = name(a1)
    assert a0 == a1 == a2


def test_name_empty_string():
    """Test that providing an empty string to name constructor raises error."""
    with pytest.raises(ValueError):
        name("")


def test_name_starts_with_digits():
    """Test that providing a string to name constructor starting with digits raises error."""
    with pytest.raises(ValueError):
        name("123")


@pytest.mark.parametrize("keyword", TEXT_SYMBOLS)
def test_name_is_a_keyword(keyword):
    """Test that parse_name with keywords as input raises error."""
    with pytest.raises(
        PDDLValidationError, match=f"invalid name '{keyword}': it is a keyword"
    ):
        parse_name(keyword)


@pytest.mark.parametrize("keyword", TEXT_SYMBOLS - {Symbols.OBJECT.value})
def test_type_is_a_keyword(keyword):
    """Test that parse_type with keywords as input raises error."""
    with pytest.raises(
        PDDLValidationError, match=f"invalid type '{keyword}': it is a keyword"
    ):
        parse_type(keyword)


def test_object_is_a_valid_type_name():
    """Test that parse_type with input 'object' does not raise error."""
    parse_type(Symbols.OBJECT.value)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Custom Exception classes."""


class PDDLError(Exception):
    """Base class for PDDL error."""


class PDDLValidationError(PDDLError):
    """Raised for PDDL error regarding validation of PDDL data structures."""


class PDDLParsingError(PDDLError):
    """Raised for PDDL parsing error."""

    def __init__(self, message: str = "Parsing Error"):
        """Initialize the PDDL parsing error exception."""
        self.message = message
        super().__init__(self.message)


class PDDLMissingRequirementError(PDDLParsingError):
    """Raised for PDDL missing requirement error."""

    def __init__(self, requirement: "Requirements"):  # type: ignore # noqa: F821
        """Initialize the PDDL missing requirement error exception."""
        self.requirement = requirement
        self.message = "Missing PDDL requirement"
        super().__init__(self.message)

    def __str__(self):
        """Get the string representation."""
        return f"{self.message}, {self.requirement} not found."



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains the definition of the PDDL requirements."""

import functools
from enum import Enum
from typing import AbstractSet, Set

from pddl.parser.symbols import RequirementSymbols as RS


@functools.total_ordering
class Requirements(Enum):
    """Enum class for the requirements."""

    STRIPS = RS.STRIPS.strip()
    TYPING = RS.TYPING.strip()
    NEG_PRECONDITION = RS.NEG_PRECONDITION.strip()
    DIS_PRECONDITION = RS.DIS_PRECONDITION.strip()
    UNIVERSAL_PRECONDITION = RS.UNIVERSAL_PRECONDITION.strip()
    EXISTENTIAL_PRECONDITION = RS.EXISTENTIAL_PRECONDITION.strip()
    QUANTIFIED_PRECONDITION = RS.QUANTIFIED_PRECONDITION.strip()
    EQUALITY = RS.EQUALITY.strip()
    CONDITIONAL_EFFECTS = RS.CONDITIONAL_EFFECTS.strip()
    ADL = RS.ADL.strip()
    DERIVED_PREDICATES = RS.DERIVED_PREDICATES.strip()
    NON_DETERMINISTIC = RS.NON_DETERMINISTIC.strip()
    FLUENTS = RS.FLUENTS.strip()
    OBJECT_FLUENTS = RS.OBJECT_FLUENTS.strip()
    NUMERIC_FLUENTS = RS.NUMERIC_FLUENTS.strip()
    ACTION_COSTS = RS.ACTION_COSTS.strip()

    @classmethod
    def quantified_precondition_requirements(cls) -> Set["Requirements"]:
        """Get the quantified precondition requirements."""
        return {
            Requirements.UNIVERSAL_PRECONDITION,
            Requirements.EXISTENTIAL_PRECONDITION,
        }

    @classmethod
    def adl_requirements(cls) -> Set["Requirements"]:
        """Get the ADL requirements."""
        return {
            Requirements.STRIPS,
            Requirements.TYPING,
            Requirements.NEG_PRECONDITION,
            Requirements.DIS_PRECONDITION,
            Requirements.EQUALITY,
            Requirements.CONDITIONAL_EFFECTS,
        }.union(cls.quantified_precondition_requirements())

    @classmethod
    def fluents_requirements(cls) -> Set["Requirements"]:
        """Get the fluents requirements."""
        return {
            Requirements.OBJECT_FLUENTS,
            Requirements.NUMERIC_FLUENTS,
        }

    def __str__(self) -> str:
        """Get the string representation."""
        return f":{self.value}"

    def __repr__(self) -> str:
        """Get an unambiguous representation."""
        return f"Requirements.{self.name}"

    def __lt__(self, other):
        """Compare with another object."""
        if isinstance(other, Requirements):
            return self.value <= other.value
        else:
            return super().__lt__(other)


def _extend_domain_requirements(
    requirements: AbstractSet[Requirements],
) -> Set[Requirements]:
    """Extend the requirements with the domain requirements."""
    extended_requirements = set(requirements)
    if Requirements.QUANTIFIED_PRECONDITION in requirements:
        extended_requirements.update(
            Requirements.quantified_precondition_requirements()
        )
    if Requirements.ADL in requirements:
        extended_requirements.update(Requirements.adl_requirements())
    if Requirements.FLUENTS in requirements:
        extended_requirements.update(Requirements.fluents_requirements())
    return extended_requirements



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Implementation of the PDDL problem parser."""

from typing import Any, Dict

from lark import ParseError, Transformer

from pddl.core import Problem
from pddl.exceptions import PDDLParsingError
from pddl.helpers.base import assert_
from pddl.logic.base import Not
from pddl.logic.functions import Divide
from pddl.logic.functions import EqualTo as FunctionEqualTo
from pddl.logic.functions import (
    Metric,
    Minus,
    NumericFunction,
    NumericValue,
    Plus,
    Times,
)
from pddl.logic.predicates import EqualTo, Predicate
from pddl.logic.terms import Constant
from pddl.parser.base import BaseParser
from pddl.parser.domain import DomainTransformer
from pddl.parser.symbols import Symbols
from pddl.requirements import Requirements


class ProblemTransformer(Transformer[Any, Problem]):
    """Problem Transformer."""

    def __init__(self) -> None:
        """Initialize the problem transformer."""
        super().__init__()

        self._domain_transformer = DomainTransformer()
        self._objects_by_name: Dict[str, Constant] = {}

    def start(self, args):
        """Process the rule 'start'."""
        return args[0]

    def problem(self, args):
        """Process the 'problem' rule."""
        args = [arg for arg in args if arg is not None]
        assert_(
            (args[0].value + args[1].value + args[-1].value == "(define)"),
            "Problem should start with '(define' and close with ')'",
        )
        return Problem(**dict(args[2:-1]))

    def problem_def(self, args):
        """Process the 'problem_def' rule."""
        return "name", args[2]

    def problem_domain(self, args):
        """Process the 'problem_domain' rule."""
        return "domain_name", args[2]

    def requirements(self, args):
        """Process the 'requirements' rule."""
        return "requirements", {Requirements(r[1:]) for r in args[2:-1]}

    def objects(self, args):
        """Process the 'problem_domain' rule."""
        object_names = args[2]
        self._objects_by_name = {
            name: Constant(name, type_tag=type_) for name, type_ in object_names.items()
        }
        return "objects", list(self._objects_by_name.values())

    def typed_list_name(self, args):
        """
        Process the 'typed_list_name' rule.

        Return a dictionary with as keys the names and as value a set of types for each name.

        :param args: the argument of this grammar rule
        :return: a typed list (name)
        """
        return self._domain_transformer.typed_list_name(args)

    def domain__type_def(self, names):
        """Process a domain type def."""
        assert_(len(names) == 1)
        return str(names[0])

    def init(self, args):
        """Process the 'init' rule."""
        flat_args = [
            item
            for sublist in args[2:-1]
            for item in (sublist if isinstance(sublist, list) else [sublist])
        ]
        return "init", flat_args

    def init_el(self, args):
        """Process the 'init_el' rule."""
        if len(args) == 1:
            return args[0]
        elif args[1] == Symbols.EQUAL.value:
            if isinstance(args[2], list) and len(args[2]) == 1:
                return FunctionEqualTo(*args[2], NumericValue(args[3]))
            elif not isinstance(args[2], list):
                return FunctionEqualTo(args[2], NumericValue(args[3]))
            else:
                funcs = [FunctionEqualTo(x, NumericValue(args[3])) for x in args[2]]
                return funcs

    def literal_name(self, args):
        """Process the 'literal_name' rule."""
        if len(args) == 1:
            return args[0]
        elif args[1] == Symbols.NOT.value:
            return Not(args[2])
        else:
            raise ParseError

    def basic_function_term(self, args):
        """Process the 'basic_function_term' rule."""
        if len(args) == 1:
            return NumericFunction(args[0])
        function_name = args[1]
        objects = [Constant(x) for x in args[2:-1]]
        return NumericFunction(function_name, *objects)

    def goal(self, args):
        """Process the 'goal' rule."""
        return "goal", args[2]

    def gd(self, args):
        """Process the 'gd_name' rule."""
        result = self._domain_transformer.gd(args)
        return result

    def atomic_formula_name(self, args):
        """Process the 'atomic_formula_name' rule."""
        if args[1] == Symbols.EQUAL.value:
            obj1 = self._objects_by_name.get(args[1])
            obj2 = self._objects_by_name.get(args[2])
            return EqualTo(obj1, obj2)
        else:
            name = args[1]
            terms = [
                (
                    Constant(str(_term_name))
                    if self._objects_by_name.get(str(_term_name)) is None
                    else self._objects_by_name.get(str(_term_name))
                )
                for _term_name in args[2:-1]
            ]
            return Predicate(name, *terms)

    def num_literal(self, args):
        """Process the 'num_literal' rule."""
        return self._domain_transformer.num_literal(args)

    def f_exp(self, args):
        """Process the 'f_exp' rule."""
        return self._domain_transformer.f_exp(args)

    def f_head(self, args):
        """Process the 'f_head' rule."""
        return self._domain_transformer.f_head(args)

    def metric_spec(self, args):
        """Process the 'metric_spec' rule."""
        if args[2] == Symbols.MINIMIZE.value:
            return "metric", Metric(args[3], args[2])
        elif args[2] == Symbols.MAXIMIZE.value:
            return "metric", Metric(args[3], args[2])
        else:
            raise PDDLParsingError(f"Unknown metric operator: {args[2]}")

    def metric_f_exp(self, args):
        """Process the 'metric_f_exp' rule."""
        if len(args) == 1:
            if not isinstance(args[0], NumericFunction):
                return NumericValue(args[0])
            return args[0]
        op = None
        if args[1] == Symbols.MINUS.value:
            op = Minus
        if args[1] == Symbols.PLUS.value:
            op = Plus
        if args[1] == Symbols.TIMES.value:
            op = Times
        if args[1] == Symbols.DIVIDE.value:
            op = Divide
        return (
            op(*args[2:-1])
            if op is not None
            else PDDLParsingError("Operator not recognized")
        )

    def atomic_formula_term(self, args):
        """Parse an atomic formula term."""
        return self._domain_transformer.atomic_formula_term(args)

    def constant(self, args):
        """Process the 'constant' rule."""
        assert_(len(args) == 1, "Unexpected parsing error.")
        return Constant(args[0])


class ProblemParser(BaseParser[Problem]):
    """PDDL problem parser class."""

    start_symbol = "problem"
    transformer_cls = ProblemTransformer



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#
"""This module contains tests for PDDL function operators."""

from pddl.logic import variables
from pddl.logic.functions import (
    EqualTo,
    GreaterEqualThan,
    GreaterThan,
    LesserEqualThan,
    LesserThan,
    NumericFunction,
)
from pddl.parser.symbols import Symbols


class TestFunctionEqualTo:
    """Test the function equal to."""

    def setup_method(self):
        """Set up the tests."""
        x, y = variables("x y", types=["type1"])
        z = variables("z", types=["type2"])
        self.function_1 = NumericFunction("function_1", x, y)
        self.function_2 = NumericFunction("function_2", *z)
        self.equal_to = EqualTo(self.function_1, self.function_2)

    def test_function(self):
        """Test the function getter."""
        assert self.equal_to.operands == (self.function_1, self.function_2)

    def test_symbol(self):
        """Test the symbol getter."""
        assert self.equal_to.SYMBOL == Symbols.EQUAL

    def test_str(self):
        """Test the str operator."""
        assert (
            str(self.equal_to)
            == f"({self.equal_to.SYMBOL.value} {self.function_1} {self.function_2})"
        )

    def test_repr(self):
        """Test the repr operator."""
        assert (
            repr(self.equal_to)
            == f"EqualTo(({repr(self.function_1)}, {repr(self.function_2)}))"
        )

    def test_equal(self):
        """Test the equal operator."""
        other = EqualTo(self.function_1, self.function_2)
        assert self.equal_to == other


class TestFunctionLesser:
    """Test the function lesser than."""

    def setup_method(self):
        """Set up the tests."""
        x, y = variables("x y", types=["type1"])
        z = variables("z", types=["type2"])
        self.function_1 = NumericFunction("function_1", x, y)
        self.function_2 = NumericFunction("function_2", *z)
        self.lesser = LesserThan(self.function_1, self.function_2)

    def test_function(self):
        """Test the function getter."""
        assert self.lesser.operands == (self.function_1, self.function_2)

    def test_symbol(self):
        """Test the symbol getter."""
        assert self.lesser.SYMBOL == Symbols.LESSER

    def test_str(self):
        """Test the str operator."""
        assert (
            str(self.lesser)
            == f"({self.lesser.SYMBOL.value} {self.function_1} {self.function_2})"
        )

    def test_repr(self):
        """Test the repr operator."""
        assert (
            repr(self.lesser)
            == f"LesserThan(({repr(self.function_1)}, {repr(self.function_2)}))"
        )

    def test_equal(self):
        """Test the equal operator."""
        other = LesserThan(self.function_1, self.function_2)
        assert self.lesser == other


class TestFunctionLesserOrEqual:
    """Test the function lesser or equal than."""

    def setup_method(self):
        """Set up the tests."""
        x, y = variables("x y", types=["type1"])
        z = variables("z", types=["type2"])
        self.function_1 = NumericFunction("function_1", x, y)
        self.function_2 = NumericFunction("function_2", *z)
        self.lesser_or_equal = LesserEqualThan(self.function_1, self.function_2)

    def test_function(self):
        """Test the function getter."""
        assert self.lesser_or_equal.operands == (self.function_1, self.function_2)

    def test_symbol(self):
        """Test the symbol getter."""
        assert self.lesser_or_equal.SYMBOL == Symbols.LESSER_EQUAL

    def test_str(self):
        """Test the str operator."""
        assert (
            str(self.lesser_or_equal)
            == f"({self.lesser_or_equal.SYMBOL.value} {self.function_1} {self.function_2})"
        )

    def test_repr(self):
        """Test the repr operator."""
        assert (
            repr(self.lesser_or_equal)
            == f"LesserEqualThan(({repr(self.function_1)}, {repr(self.function_2)}))"
        )

    def test_equal(self):
        """Test the equal operator."""
        other = LesserEqualThan(self.function_1, self.function_2)
        assert self.lesser_or_equal == other


class TestFunctionGreater:
    """Test the function greater than."""

    def setup_method(self):
        """Set up the tests."""
        x, y = variables("x y", types=["type1"])
        z = variables("z", types=["type2"])
        self.function_1 = NumericFunction("function_1", x, y)
        self.function_2 = NumericFunction("function_2", *z)
        self.greater = GreaterThan(self.function_1, self.function_2)

    def test_function(self):
        """Test the function getter."""
        assert self.greater.operands == (self.function_1, self.function_2)

    def test_symbol(self):
        """Test the symbol getter."""
        assert self.greater.SYMBOL == Symbols.GREATER

    def test_str(self):
        """Test the str operator."""
        assert (
            str(self.greater)
            == f"({self.greater.SYMBOL.value} {self.function_1} {self.function_2})"
        )

    def test_repr(self):
        """Test the repr operator."""
        assert (
            repr(self.greater)
            == f"GreaterThan(({repr(self.function_1)}, {repr(self.function_2)}))"
        )

    def test_equal(self):
        """Test the equal operator."""
        other = GreaterThan(self.function_1, self.function_2)
        assert self.greater == other


class TestFunctionGreaterOrEqual:
    """Test the function greater or equal than."""

    def setup_method(self):
        """Set up the tests."""
        x, y = variables("x y", types=["type1"])
        z = variables("z", types=["type2"])
        self.function_1 = NumericFunction("function_1", x, y)
        self.function_2 = NumericFunction("function_2", *z)
        self.greater_or_equal = GreaterEqualThan(self.function_1, self.function_2)

    def test_function(self):
        """Test the function getter."""
        assert self.greater_or_equal.operands == (self.function_1, self.function_2)

    def test_symbol(self):
        """Test the symbol getter."""
        assert self.greater_or_equal.SYMBOL == Symbols.GREATER_EQUAL

    def test_str(self):
        """Test the str operator."""
        assert (
            str(self.greater_or_equal)
            == f"({self.greater_or_equal.SYMBOL.value} {self.function_1} {self.function_2})"
        )

    def test_repr(self):
        """Test the repr operator."""
        assert (
            repr(self.greater_or_equal)
            == f"GreaterEqualThan(({repr(self.function_1)}, {repr(self.function_2)}))"
        )

    def test_equal(self):
        """Test the equal operator."""
        other = GreaterEqualThan(self.function_1, self.function_2)
        assert self.greater_or_equal == other



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains the definition to deal with symbols."""

from enum import Enum
from typing import Set

OpSymbol = str
OpRequirement = str


class Symbols(Enum):
    """A set of symbols that can be used in PDDL."""

    ROUND_BRACKET_LEFT = "("
    ROUND_BRACKET_RIGHT = ")"
    TYPE_SEP = "-"
    EQUAL = "="
    ACTION = ":action"
    AND = "and"
    CONSTANTS = ":constants"
    DEFINE = "define"
    DERIVED = ":derived"
    DOMAIN = "domain"
    DOMAIN_P = ":domain"
    EFFECT = ":effect"
    EITHER = "either"
    EXISTS = "exists"
    FORALL = "forall"
    GOAL = ":goal"
    IMPLY = "imply"
    INIT = ":init"
    NOT = "not"
    OBJECT = "object"
    OBJECTS = ":objects"
    ONEOF = "oneof"
    OR = "or"
    PARAMETERS = ":parameters"
    PRECONDITION = ":precondition"
    PREDICATES = ":predicates"
    PROBLEM = "problem"
    REQUIREMENTS = ":requirements"
    TYPES = ":types"
    METRIC = ":metric"
    WHEN = "when"
    GREATER_EQUAL = ">="
    GREATER = ">"
    LESSER_EQUAL = "<="
    LESSER = "<"
    MINUS = "-"
    PLUS = "+"
    TIMES = "*"
    DIVIDE = "/"
    ASSIGN = "assign"
    SCALE_UP = "scale-up"
    SCALE_DOWN = "scale-down"
    INCREASE = "increase"
    DECREASE = "decrease"
    MAXIMIZE = "maximize"
    MINIMIZE = "minimize"
    TOTAL_COST = "total-cost"


ALL_SYMBOLS: Set[str] = {v.value for v in Symbols}
BINARY_COMP_SYMBOLS: Set[str] = {
    v.value for v in Symbols if v.value in {">=", ">", "<=", "<", "="}
}


class RequirementSymbols(Enum):
    """A set of requirements that can be used in PDDL."""

    STRIPS = ":strips"
    TYPING = ":typing"
    NEG_PRECONDITION = ":negative-preconditions"
    DIS_PRECONDITION = ":disjunctive-preconditions"
    UNIVERSAL_PRECONDITION = ":universal-preconditions"
    EXISTENTIAL_PRECONDITION = ":existential-preconditions"
    QUANTIFIED_PRECONDITION = ":quantified-preconditions"
    EQUALITY = ":equality"
    CONDITIONAL_EFFECTS = ":conditional-effects"
    ADL = ":adl"
    DERIVED_PREDICATES = ":derived-predicates"
    NON_DETERMINISTIC = ":non-deterministic"
    FLUENTS = ":fluents"
    OBJECT_FLUENTS = ":object-fluents"
    NUMERIC_FLUENTS = ":numeric-fluents"
    ACTION_COSTS = ":action-costs"

    def strip(self) -> str:
        """Strip the leading colon."""
        return self.value[1:]


ALL_REQUIREMENTS: Set[str] = {v.value for v in RequirementSymbols}



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This test module contains the fixtures for 'blocksworld-ipc08' domain and problem."""

import pytest

from pddl.action import Action
from pddl.core import Domain, Problem
from pddl.logic.base import And, OneOf
from pddl.logic.helpers import constants, variables
from pddl.logic.predicates import EqualTo, Predicate
from pddl.requirements import Requirements


@pytest.fixture(scope="session")
def blocksworld_domain():
    """The 'blocksworld' domain."""
    # types
    block = "block"

    # terms
    b, b1, b2, b3 = variables("b b1 b2 b3", types=[block])

    # constants:
    constants = None

    # predicates
    holding = Predicate("holding", b)
    emptyhand = Predicate("emptyhand")
    on_table = Predicate("on-table", b)
    on = Predicate("on", b1, b2)
    clear = Predicate("clear", b)
    predicates = {holding, emptyhand, on_table, on, clear}

    # actions
    # pick-up
    pick_up_name = "pick-up"
    pick_up_parameters = [b1, b2]
    pick_up_precondition = ~EqualTo(b1, b2) & emptyhand & clear(b1) & on(b1, b2)
    pick_up_effect = OneOf(
        And(holding(b1), clear(b2), ~emptyhand, ~clear(b1), ~on(b1, b2)),
        And(clear(b2), on_table(b1), ~on(b1, b2)),
    )
    pick_up = Action(
        pick_up_name, pick_up_parameters, pick_up_precondition, pick_up_effect
    )

    # pick-up-from-table
    pick_up_from_table_name = "pick-up-from-table"
    pick_up_from_table_parameters = [b]
    pick_up_from_table_precondition = emptyhand & clear(b) & on_table(b)
    pick_up_from_table_effect = OneOf(And(), And(holding(b), ~emptyhand, ~on_table(b)))
    pick_up_from_table = Action(
        pick_up_from_table_name,
        pick_up_from_table_parameters,
        pick_up_from_table_precondition,
        pick_up_from_table_effect,
    )

    # put-on-block
    put_on_block_name = "put-on-block"
    put_on_block_parameters = [b1, b2]
    put_on_block_precondition = holding(b1) & clear(b2)
    put_on_block_effect = OneOf(
        And(on(b1, b2), emptyhand, clear(b1), ~holding(b1), ~clear(b2)),
        And(on_table(b1), emptyhand, clear(b1), ~holding(b1)),
    )
    put_on_block = Action(
        put_on_block_name,
        put_on_block_parameters,
        put_on_block_precondition,
        put_on_block_effect,
    )

    # put-down
    put_down_name = "put-down"
    put_down_parameters = [b]
    put_down_precondition = holding(b)
    put_down_effect = And(on_table(b), emptyhand, clear(b), ~holding(b))
    put_down = Action(
        put_down_name, put_down_parameters, put_down_precondition, put_down_effect
    )
    # pick-tower
    pick_tower_name = "pick-tower"
    pick_tower_parameters = [b1, b2, b3]
    pick_tower_precondition = emptyhand & on(b1, b2) & on(b2, b3)
    pick_tower_effect = OneOf(
        And(), And(holding(b2), clear(b3), ~emptyhand, ~on(b2, b3))
    )
    pick_tower = Action(
        pick_tower_name,
        pick_tower_parameters,
        pick_tower_precondition,
        pick_tower_effect,
    )

    # put-tower-on-block
    put_tower_on_block_name = "put-tower-on-block"
    put_tower_on_block_parameters = [b1, b2, b3]
    put_tower_on_block_precondition = holding(b2) & on(b1, b2) & clear(b3)
    put_tower_on_block_effect = OneOf(
        And(on(b2, b3), emptyhand, ~holding(b2), ~clear(b3)),
        And(on_table(b2), emptyhand, ~holding(b2)),
    )
    put_tower_on_block = Action(
        put_tower_on_block_name,
        put_tower_on_block_parameters,
        put_tower_on_block_precondition,
        put_tower_on_block_effect,
    )

    # put-tower-down
    put_tower_down_name = "put-tower-down"
    put_tower_down_parameters = [b1, b2]
    put_tower_down_precondition = holding(b2) & on(b1, b2)
    put_tower_down_effect = And(on_table(b2), emptyhand, ~holding(b2))
    put_tower_down = Action(
        put_tower_down_name,
        put_tower_down_parameters,
        put_tower_down_precondition,
        put_tower_down_effect,
    )

    name = "blocks-domain"
    requirements = {
        Requirements.EQUALITY,
        Requirements.NON_DETERMINISTIC,
        Requirements.TYPING,
        Requirements.NEG_PRECONDITION,
    }
    types = dict.fromkeys([block])
    actions = {
        pick_up,
        pick_up_from_table,
        put_on_block,
        put_down,
        pick_tower,
        put_tower_on_block,
        put_tower_down,
    }
    domain = Domain(
        name=name,
        requirements=requirements,
        types=types,
        constants=constants,
        predicates=predicates,
        actions=actions,
    )
    return domain


@pytest.fixture(scope="session")
def blocksworld_problem_01():
    """Blocksworld ipc08 problem 01."""
    # objects
    objects = [b1, b2, b3, b4, b5] = constants("b1 b2 b3 b4 b5", type_="block")

    # predicates
    emptyhand = Predicate("emptyhand")
    on_table = Predicate("on-table", b1)
    on = Predicate("on", b1, b2)
    clear = Predicate("clear", b1)

    init = {
        emptyhand,
        on(b1, b3),
        on(b2, b1),
        on_table(b3),
        on_table(b4),
        on(b5, b4),
        clear(b2),
        clear(b5),
    }

    goal = (
        emptyhand
        & on(b1, b2)
        & on(b2, b5)
        & on_table(b3)
        & on_table(b4)
        & on_table(b5)
        & clear(b1)
        & clear(b3)
        & clear(b4)
    )

    problem_name = "bw_5_1"

    problem = Problem(
        problem_name,
        domain_name="blocks-domain",
        objects=objects,
        init=init,
        goal=goal,
    )
    return problem



#!/usr/bin/env python3
#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains tests verifying that the formatter outputs are syntactically valid and semantically faithful."""

from pathlib import Path
from textwrap import dedent

import pytest

from pddl.action import Action
from pddl.core import Domain, Problem
from pddl.formatter import domain_to_string, problem_to_string
from pddl.logic import Constant, Variable, constants
from pddl.logic.base import And, ForallCondition
from pddl.logic.functions import (
    EqualTo,
    GreaterEqualThan,
    Increase,
    LesserEqualThan,
    NumericFunction,
    NumericValue,
)
from pddl.requirements import Requirements
from tests.conftest import DOMAIN_FILES, PROBLEM_FILES


@pytest.mark.parametrize("pddl_file", DOMAIN_FILES)
def test_domain_formatter(domain_parser, pddl_file: Path):
    """Test generated domain formatting."""
    expected_domain_obj = domain_parser(pddl_file.read_text())
    actual_domain_str = domain_to_string(expected_domain_obj)
    actual_domain_obj = domain_parser(actual_domain_str)
    assert (
        actual_domain_obj == expected_domain_obj
    ), f"Domain mismatch for {str(pddl_file).split('/')[-2]}"


@pytest.mark.parametrize("pddl_file", PROBLEM_FILES)
def test_problem_formatter(problem_parser, pddl_file):
    """Test generated problem formatting."""
    expected_problem_obj = problem_parser(pddl_file.read_text())
    actual_problem_str = problem_to_string(expected_problem_obj)
    actual_problem_obj = problem_parser(actual_problem_str)
    assert actual_problem_obj == expected_problem_obj


def test_typed_constants_formatting_in_domain() -> None:
    """Test that types and typed constants are formatted correctly."""
    t1, t2, t3 = "type_1", "type_2", "type_3"

    a1, b1, c1 = constants("a b c", type_=t1)
    d2, e2, f2 = constants("d e f", type_=t2)
    g3, h3, i3 = constants("g h i", type_=t3)
    j, k, lc = constants("j k l", type_=None)

    # define the domain object.
    domain = Domain(
        "my_domain",
        requirements=[Requirements.TYPING],
        types={t1: None, t2: t1, t3: t1},
        constants=[a1, b1, c1, d2, e2, f2, g3, h3, i3, j, k, lc],
    )

    domain_str = domain_to_string(domain)

    assert domain_str == dedent("""\
    (define (domain my_domain)
        (:requirements :typing)
        (:types
            type_1 - object
            type_2 type_3 - type_1
        )
        (:constants a b c - type_1 d e f - type_2 g h i - type_3 j k l)
    )""")


def test_typed_objects_formatting_in_problem() -> None:
    """Test that typed objects are formatted correctly."""
    t1, t2, t3 = "type_1", "type_2", "type_3"

    a1, b1, c1 = constants("a b c", type_=t1)
    d2, e2, f2 = constants("d e f", type_=t2)
    g3, h3, i3 = constants("g h i", type_=t3)
    j, k, lc = constants("j k l", type_=None)

    problem = Problem(
        "problem-1",
        domain_name="my_domain",
        requirements=[Requirements.TYPING],
        objects=[a1, b1, c1, d2, e2, f2, g3, h3, i3, j, k, lc],
    )
    problem_str = problem_to_string(problem)

    assert problem_str == dedent("""\
    (define (problem problem-1)
        (:domain my_domain)
        (:requirements :typing)
        (:objects a b c - type_1 d e f - type_2 g h i - type_3 j k l)
        (:init )
        (:goal (and ))
    )""")


def test_numerical_hello_world_domain_formatter():
    """Test that numerical NumericFunctions are formatted correctly."""
    neighbor = Variable("neighbor")
    hello_counter = NumericFunction("hello_counter", neighbor)
    action = Action(
        "say-hello-world",
        parameters=[neighbor],
        precondition=LesserEqualThan(hello_counter, NumericValue(3)),
        effect=And(Increase(hello_counter, NumericValue(1))),
    )

    domain = Domain(
        name="hello-world-functions",
        requirements=[Requirements.STRIPS, Requirements.NUMERIC_FLUENTS],
        functions={hello_counter: None},
        actions=[action],
    )

    assert domain_to_string(domain) == "\n".join(
        (
            "(define (domain hello-world-functions)",
            "    (:requirements :numeric-fluents :strips)",
            "    (:functions (hello_counter ?neighbor))",
            "    (:action say-hello-world",
            "        :parameters (?neighbor)",
            "        :precondition (<= (hello_counter ?neighbor) 3)",
            "        :effect (increase (hello_counter ?neighbor) 1)",
            "    )",
            ")",
        )
    )


def test_numerical_hello_world_problem_formatter():
    """Test that numerical NumericFunctions are formatted correctly."""
    neighbors = [Constant(name, "neighbor") for name in ("Alice", "Bob", "Charlie")]
    problem = Problem(
        name="hello-3-times",
        domain_name="hello-world-functions",
        objects=neighbors,
        init=[
            EqualTo(NumericFunction("hello_counter", neighbor), NumericValue(0))
            for neighbor in neighbors
        ],
        goal=ForallCondition(
            GreaterEqualThan(
                NumericFunction("hello_counter", Variable("neighbor")), NumericValue(1)
            ),
            [Variable("neighbor", ["neighbor"])],
        ),
    )

    assert problem_to_string(problem) == "\n".join(
        (
            "(define (problem hello-3-times)",
            "    (:domain hello-world-functions)",
            "    (:objects Alice Bob Charlie - neighbor)",
            "    (:init (= (hello_counter Alice) 0) (= (hello_counter Bob) 0) (= (hello_counter Charlie) 0))",
            "    (:goal (forall (?neighbor - neighbor) (>= (hello_counter ?neighbor) 1)))",
            ")",
        )
    )



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This modules implements PDDL terms."""

import functools
from typing import AbstractSet, Any, Collection, Optional

from pddl.custom_types import name as name_type
from pddl.custom_types import namelike, parse_name, to_type
from pddl.helpers.base import assert_, check_no_duplicates, ensure_set
from pddl.helpers.cache_hash import cache_hash


def _print_tag_set(type_tags: AbstractSet[Any]) -> str:
    """Print a tag set."""
    if len(type_tags) == 0:
        return "[]"
    return repr(sorted(map(str, type_tags)))


@cache_hash
@functools.total_ordering
class Term:
    """A term in a formula."""

    def __init__(
        self, term_name: namelike, type_tags: Optional[Collection[namelike]] = None
    ):
        """
        Initialize a term.

        :param term_name: the name for the term.
        :param type_tags: the type tags associated to this term.
        """
        assert_(type(self) is not Term, "Term is an abstract class")
        self._name = parse_name(term_name)
        self._type_tags = frozenset(to_type(ensure_set(check_no_duplicates(type_tags))))  # type: ignore

    @property
    def name(self) -> name_type:
        """Get the name."""
        return self._name

    @property
    def type_tags(self) -> AbstractSet[name_type]:
        """Get a set of type tags for this term."""
        return self._type_tags

    def __eq__(self, other):
        """Compare with another term."""
        return (
            isinstance(other, Term)
            and self.name == other.name
            and self.type_tags == other.type_tags
        )

    def __lt__(self, other):
        """Compare with another term."""
        if not isinstance(other, Term):
            return super().__lt__(other)
        lhs = (type(self).__name__, self.name, sorted(self.type_tags))
        rhs = (type(other).__name__, other.name, sorted(other.type_tags))
        return lhs < rhs


# TODO check correctness
class Constant(Term):
    """A constant term."""

    def __init__(self, name: namelike, type_tag: Optional[namelike] = None):
        """
        Initialize a constant.

        :param name: the name.
        :param type_tag: the type tag
        """
        super().__init__(name, type_tags={type_tag} if type_tag is not None else None)

    @property
    def type_tag(self) -> Optional[name_type]:
        """Get the type of this constant."""
        type_tags = self.type_tags
        assert_(len(type_tags) <= 1)
        return next(iter(type_tags)) if len(type_tags) == 1 else None

    def __str__(self) -> str:
        """Get the string representation."""
        return self._name

    def __repr__(self):
        """Get a unique representation of the object."""
        return f"Constant({self.name})"

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return isinstance(other, Constant) and self.name == other.name

    def __hash__(self):
        """Get the hash."""
        return hash((Constant, self._name))


class Variable(Term):
    """A variable term."""

    def __init__(
        self, name: namelike, type_tags: Optional[Collection[namelike]] = None
    ):
        """
        Initialize the variable.

        :param name: the name.
        :param type_tags: the type tags
        """
        super().__init__(name, type_tags=type_tags)

    def __str__(self) -> str:
        """Get the string representation."""
        return f"?{self._name}"

    def __repr__(self):
        """Get a unique representation of the object."""
        return f"Variable({self.name})"

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return (
            isinstance(other, Variable)
            and self.name == other.name
            and self.type_tags == other.type_tags
        )

    def __hash__(self) -> int:
        """Get the hash."""
        return hash((Variable, self._name))

    def with_type_tags(self, type_tags: Optional[Collection[namelike]]) -> "Variable":
        """Return a new variable with the given type tags."""
        return Variable(self._name, type_tags=type_tags)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This test module contains the fixtures for 'triangle-tireworld' domain and problem."""

import pytest

from pddl.action import Action
from pddl.core import Domain, Problem
from pddl.logic.base import And, OneOf
from pddl.logic.helpers import constants, variables
from pddl.logic.predicates import Predicate
from pddl.requirements import Requirements


@pytest.fixture(scope="session")
def triangle_tireworld_domain():
    """The 'triangle-tireworld' domain."""
    # types
    location = "location"

    # terms
    to, from_, loc = variables("to from loc", types=[location])

    # constants:
    constants = None

    # predicates
    vehicleat = Predicate("vehicleat", loc)
    spare_in = Predicate("spare-in", loc)
    road = Predicate("road", from_, to)
    not_flattire = Predicate("not-flattire")
    predicates = {vehicleat, spare_in, road, not_flattire}

    # actions
    # move-car
    move_car_name = "move-car"
    move_car_parameters = [from_, to]
    move_car_precondition = vehicleat(from_) & road(from_, to) & not_flattire
    move_car_effect = And(
        OneOf(
            And(vehicleat(to), ~vehicleat(from_)),
            And(vehicleat(to), ~vehicleat(from_), ~not_flattire),
        )
    )
    move_car = Action(
        move_car_name, move_car_parameters, move_car_precondition, move_car_effect
    )

    # changetire
    changetire_name = "changetire"
    changetire_parameters = [loc]
    changetire_precondition = spare_in(loc) & vehicleat(loc)
    changetire_effect = And(~spare_in(loc), not_flattire)
    changetire = Action(
        changetire_name,
        changetire_parameters,
        changetire_precondition,
        changetire_effect,
    )

    name = "triangle-tire"
    requirements = {
        Requirements.STRIPS,
        Requirements.NON_DETERMINISTIC,
        Requirements.TYPING,
    }
    types = dict.fromkeys([location])
    actions = {
        move_car,
        changetire,
    }
    domain = Domain(
        name=name,
        requirements=requirements,
        types=types,
        constants=constants,
        predicates=predicates,
        actions=actions,
    )
    return domain


@pytest.fixture(scope="session")
def triangle_tireworld_problem_01():
    """Triangle-tireworld problem 01."""
    # objects
    objects = [l1, l2, l3, l4, l5, l6, l7, l8, l9] = constants(
        "l1 l2 l3 l4 l5 l6 l7 l8 l9", type_="location"
    )

    # predicates
    vehicleat = Predicate("vehicleat", l1)
    spare_in = Predicate("spare-in", l1)
    road = Predicate("road", l1, l2)
    not_flattire = Predicate("not-flattire")

    init = {
        vehicleat(l1),
        road(l1, l2),
        road(l2, l3),
        road(l1, l4),
        road(l2, l5),
        road(l4, l2),
        road(l5, l3),
        road(l4, l7),
        road(l7, l5),
        spare_in(l4),
        spare_in(l5),
        spare_in(l7),
        spare_in(l7),
    }

    goal = vehicleat(l3)

    problem_name = "triangle-tire-1"

    problem = Problem(
        problem_name,
        domain_name="triangle-tire",
        objects=objects,
        init=init,
        goal=goal,
    )
    return problem



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains tests for PDDL actions."""

from pddl.action import Action


class TestActionSimpleInitialization:
    """Test simple action initialization."""

    def setup_method(self):
        """Set up the tests."""
        self.action = Action("action", [])

    def test_name(self):
        """Test the name getter."""
        assert self.action.name == "action"

    def test_parameters(self):
        """Test the parameters getter."""
        assert self.action.parameters == tuple()

    def test_precondition(self):
        """Test the precondition getter."""
        assert self.action.precondition is None

    def test_effects(self):
        """Test the effects getter."""
        assert self.action.effect is None



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Implementation of the PDDL plan parser."""

from typing import Any

from lark import Transformer

from pddl.core import Plan
from pddl.logic.terms import Constant
from pddl.parser.base import BaseParser


class PlanTransformer(Transformer[Any, Plan]):
    """Plan Transformer."""

    def __init__(self) -> None:
        """Initialize the plan transformer."""
        super().__init__()

    def ground_action(self, args):
        """Process the 'ground_action' rule."""
        action_name = args[1]
        if len(args) > 3:
            parameters = args[2:-1]
        else:
            parameters = []
        parameters = [Constant(str(param)) for param in parameters]
        return (action_name, parameters)

    def plan(self, args):
        """Process the 'plan' rule."""
        actions = args
        return Plan(actions=actions)


class PlanParser(BaseParser[Plan]):
    """A class for parsing PDDL plans."""

    transformer_cls = PlanTransformer
    start_symbol = "plan"



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Base classes for PDDL logic formulas."""

import functools
from abc import abstractmethod
from typing import AbstractSet, Any, Collection, List, Mapping, Optional, Sequence

from pddl.helpers.base import ensure_set
from pddl.helpers.cache_hash import cache_hash
from pddl.logic.terms import Term, Variable
from pddl.parser.symbols import Symbols


@cache_hash
class Formula:
    """Base class for all the formulas."""

    @abstractmethod
    def instantiate(self, mapping: Mapping[Variable, Term]) -> "Formula":
        """Instantiate the formula with a mapping from variables to terms."""
        raise NotImplementedError()

    def __invert__(self) -> "Formula":
        """Negate the formula."""
        return Not(self)

    def __and__(self, other: "Formula") -> "Formula":
        """Put in and with another formula."""
        return And(self, other)

    def __or__(self, other: "Formula") -> "Formula":
        """Put in or with another formula."""
        return Or(self, other)

    def __rshift__(self, other: "Formula") -> "Formula":
        """Define A implies B."""
        return Or(Not(self), other)


class BinaryOp(Formula):
    """Binary operator."""

    SYMBOL: str

    def __init__(self, *operands: Formula):
        """
        Init a binary operator.

        :param operands: the operands.
        """
        self._operands = list(operands)

    @property
    def operands(self) -> Sequence[Formula]:
        """Get the operands."""
        return tuple(self._operands)

    def instantiate(self, mapping: Mapping[Variable, Term]) -> "Formula":
        """Instantiate the formula with a mapping from variables to terms."""
        return type(self)(*(op.instantiate(mapping) for op in self.operands))

    def __str__(self) -> str:
        """Get the string representation."""
        return f"({self.SYMBOL} {' '.join(map(str, self.operands))})"

    def __repr__(self) -> str:
        """Get an unambiguous string representation."""
        return f"{type(self).__name__}({repr(self._operands)})"

    def __eq__(self, other):
        """Compare with another object."""
        return isinstance(other, type(self)) and self.operands == other.operands

    def __hash__(self) -> int:
        """Compute the hash of the object."""
        return hash((type(self), self.operands))


class UnaryOp(Formula):
    """Unary operator."""

    SYMBOL: str

    def __init__(self, arg: Formula):
        """
        Initialize the unary operator.

        :param arg: the argument.
        """
        self._arg = arg

    @property
    def argument(self) -> Formula:
        """Get the argument."""
        return self._arg

    def instantiate(self, mapping: Mapping[Variable, Term]) -> "Formula":
        """Instantiate the formula with a mapping from variables to terms."""
        return type(self)(self.argument.instantiate(mapping))

    def __str__(self) -> str:
        """Get the string representation."""
        return f"({self.SYMBOL} {self.argument})"

    def __repr__(self) -> str:
        """Get an unambiguous string representation."""
        return f"{type(self).__name__}({repr(self.argument)})"

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return isinstance(other, type(self)) and self.argument == other.argument

    def __hash__(self) -> int:
        """Compute the hash of the object."""
        return hash((type(self), self.argument))


class Atomic(Formula):
    """Atomic formula."""

    @abstractmethod
    def instantiate(self, mapping: Mapping[Variable, Term]) -> "Formula":
        """Instantiate the formula with a mapping from variables to terms."""
        raise NotImplementedError()


class BinaryOpMetaclass(type):
    """Metaclass to simplify monotone operator instantiations."""

    _absorbing: Optional[Formula] = None
    idempotency: bool = False

    def __call__(cls, *args, **kwargs):
        """Init the subclass object."""
        operands = _simplify_monotone_op_operands(
            cls, *args, idempotency=cls.idempotency
        )
        if len(operands) == 1 and cls.idempotency:
            return operands[0]

        return super(BinaryOpMetaclass, cls).__call__(*operands, **kwargs)


class And(BinaryOp, metaclass=BinaryOpMetaclass):
    """And operator."""

    _absorbing = False
    idempotency = True
    SYMBOL = "and"


class Or(BinaryOp, metaclass=BinaryOpMetaclass):
    """Or operator."""

    _absorbing = True
    idempotency = True
    SYMBOL = "or"


class OneOf(BinaryOp):
    """OneOf operator."""

    SYMBOL = "oneof"


class Imply(BinaryOp):
    """Imply operator."""

    SYMBOL = "imply"


class Not(UnaryOp):
    """Not operator."""

    SYMBOL = "not"


@cache_hash
@functools.total_ordering
class QuantifiedCondition(Formula):
    """Superclass for quantified conditions."""

    SYMBOL: str

    def __init__(
        self, cond: "Formula", variables: Optional[Collection[Variable]] = None
    ) -> None:
        """Initialize the quantified condition."""
        self._cond = cond
        self._variables = ensure_set(variables)

    @property
    def condition(self) -> "Formula":
        """Get the condition."""
        return self._cond

    @property
    def variables(self) -> AbstractSet[Variable]:
        """Get the variables."""
        return self._variables

    def instantiate(self, mapping: Mapping[Variable, Term]) -> "Formula":
        """Instantiate the formula with a mapping from variables to terms."""
        return type(self)(self.condition.instantiate(mapping), self.variables)

    def __str__(self) -> str:
        """Get the string representation."""

        def build_tags(tags):
            if len(tags) == 0:
                return ""
            return f" - {' '.join(tags)}"

        var_block = " ".join([f"{v}{build_tags(v.type_tags)}" for v in self.variables])
        return f"({self.SYMBOL} ({var_block}) {self.condition})"

    def __repr__(self) -> str:
        """Get an unambiguous string representation."""
        return f"{type(self).__name__}({self.variables}, {self.condition})"

    def __eq__(self, other) -> bool:
        """Compare with another object."""
        return (
            isinstance(other, type(self))
            and self.variables == other.variables
            and self.condition == other.condition
        )

    def __hash__(self) -> int:
        """Compute the hash of the object."""
        return hash((type(self), self.variables, self.condition))

    def __lt__(self, other):
        """Compare with another object."""
        if isinstance(other, QuantifiedCondition):
            return (self.variables, self.condition) < (other.variables, other.condition)
        return super().__lt__(other)


class ForallCondition(QuantifiedCondition):
    """Forall Condition."""

    SYMBOL = Symbols.FORALL.value


class ExistsCondition(QuantifiedCondition):
    """Exists Condition."""

    SYMBOL = Symbols.EXISTS.value


def ensure_formula(f: Optional[Formula], is_none_true: bool) -> Formula:
    """
    Ensure the argument is a formula.

    :param f: the formula, or None.
    :param is_none_true: if true, None reduces to And(); FalseFormula otherwise.
    :return: the same set, or an empty set if the arg was None.
    """
    return f if f is not None else And() if is_none_true else Or()


def is_literal(formula: Formula) -> bool:
    """
    Check whether a formula is a literal.

    That is, whether it is one of the following:
    - an atomic formula,
    - a Not formula whose argument is an atomic formula.

    :param formula: the formula.
    :return: True if the formula is a literal; False otherwise.
    """
    return (
        isinstance(formula, Atomic)
        or isinstance(formula, Not)
        and isinstance(formula.argument, Atomic)
    )


def _simplify_monotone_op_operands(cls, *operands, idempotency: bool = False):
    old_operands: List[Any] = (
        list(dict.fromkeys(operands)) if idempotency else list(operands)
    )
    if len(old_operands) == 0:
        return []
    elif len(old_operands) == 1:
        return [old_operands[0]]

    # shift-up subformulas with same operator. DFS on expression tree.
    seen = set()
    new_operands = []
    stack = old_operands[::-1]  # it is reversed in order to preserve order.
    while len(stack) > 0:
        element = stack.pop()
        if not isinstance(element, cls):
            if element not in seen:
                seen.add(element)
                new_operands.append(element)
            continue
        stack.extend(reversed(element.operands))  # see above regarding reversed.

    return new_operands



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Top-level package for pddl."""

from .__version__ import (
    __author__,
    __author_email__,
    __copyright__,
    __description__,
    __license__,
    __title__,
    __url__,
    __version__,
)
from .helpers.base import _get_current_path

_ROOT_PATH = _get_current_path()


# Simple helpers
def parse_domain(fn):
    from pddl.parser.domain import DomainParser

    with open(fn, "r") as f:
        dtext = f.read()
    return DomainParser()(dtext)


def parse_problem(fn):
    from pddl.parser.problem import ProblemParser

    with open(fn, "r") as f:
        ptext = f.read()
    return ProblemParser()(ptext)


def parse_plan(fn):
    from pddl.parser.plan import PlanParser

    with open(fn, "r") as f:
        plantext = f.read()
    return PlanParser()(plantext)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains tests for the CLI tool."""

import logging
from pathlib import Path

import pytest
from click.testing import CliRunner

from pddl.__main__ import cli
from tests.conftest import DOMAIN_FILES, PLAN_FILES, PROBLEM_FILES


@pytest.mark.parametrize("pddl_file", DOMAIN_FILES)
def test_pddl_domain_cli(pddl_file):
    """Test PDDL domain-only CLI invocation."""
    logging.debug(f"Parsing domain file: {pddl_file}")
    result = CliRunner().invoke(cli, args=[str(pddl_file)], catch_exceptions=False)
    assert result.exit_code == 0


def _find_domain_for_problem(problem_file: Path) -> Path:
    """Find the nearest domain.pddl for a given problem file."""
    for parent in [problem_file.parent, *problem_file.parents]:
        domain_file = parent / "domain.pddl"
        if domain_file.exists():
            return domain_file
    raise RuntimeError(f"Could not find domain.pddl for problem file {problem_file}")


def _find_problem_for_plan(plan_file: Path) -> Path:
    """Find the matching problem file for a given plan file."""
    problem_file = plan_file.with_suffix(".pddl")
    if problem_file.exists():
        return problem_file
    raise RuntimeError(f"Could not find matching problem file for plan {plan_file}")


@pytest.mark.parametrize("problem_file", PROBLEM_FILES)
def test_pddl_problem_cli(problem_file):
    """Test PDDL domain+problem CLI invocation."""
    domain_file = _find_domain_for_problem(problem_file)
    logging.debug(f"Parsing problem file: {problem_file} with domain {domain_file}")
    result = CliRunner().invoke(
        cli, args=[str(domain_file), str(problem_file)], catch_exceptions=False
    )
    assert result.exit_code == 0


@pytest.mark.parametrize("plan_file", PLAN_FILES)
def test_pddl_plan_cli(plan_file):
    """Test PDDL domain+problem+plan CLI invocation."""
    problem_file = _find_problem_for_plan(plan_file)
    domain_file = _find_domain_for_problem(problem_file)
    logging.debug(
        f"Parsing plan file: {plan_file} with problem {problem_file} and domain {domain_file}"
    )
    result = CliRunner().invoke(
        cli,
        args=[str(domain_file), str(problem_file), str(plan_file)],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
